# Copyright (c) 2022 Valio
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT



import typing
from dataclasses import astuple, dataclass, fields
from pprint import pformat
from typing import Any

from typingx import issubclassx
from valio.error import errors
from valio.logger import loggers

__all__ = ["Property", "NAME", "DEFAULT", "DOC", "DEBUG"]

T = typing.TypeVar('T')
PropBound = typing.TypeVar('PropBound', bound=typing.Union['Property', None])

Union = typing.Union[T, PropBound]

NAME = Union[str, PropBound]
DEFAULT = Union[Any, PropBound]
DOC = Union[str, PropBound]
DEBUG = Union[bool, PropBound]


@dataclass
class Property(loggers.Logger):
    """Property Class: its a base class for all property related usages.
    it has an inbuilt support for logging for each and every activity that
    the class does.

    Usage:

    >>> from dataclasses import dataclass
    ...
    >>> Name: str = Property()
    >>> Salary: int = Property()
    ...
    ...
    >>> @dataclass
    ... class Persons(object):
    ...     name: str = Name
    ...     salary: int = Salary
    ...
    >>> p = Persons()
    >>> p.name = "some name"
    >>> p.salary = "1000000"
    ...
    """

    name: NAME = None
    default: DEFAULT = None
    doc: DOC = None
    debug: DEBUG = None
    errors: Union[list, None] = None

    def __init__(
            self,
            name: NAME = None,
            default: DEFAULT = None,
            doc: DOC = None,
            debug: DEBUG = None,
            **kwargs,
    ):
        if name is None or isinstance(name, str):
            self.name = name
        else:
            raise TypeError(
                f"name expected type {str.__name__} value, "
                f"got {type(name).__name__} type instead"
            )

        if doc is None or isinstance(doc, str):
            self.doc = doc
        else:
            raise TypeError(
                f"doc expected type {str.__name__} value, "
                f"got {type(doc).__name__} type instead"
            )

        if debug is None or isinstance(debug, bool):
            self.debug = debug  # or environments.DEBUG
        else:
            raise TypeError(
                f"debug expected type {bool.__name__} value, "
                f"got {type(debug).__name__} type instead"
            )
        self.default = default
        self.annotation = getattr(self, "annotation", None)
        self._dict = None

        super(Property, self).__init__(**kwargs)

    def pre_set(self, obj, value):
        """All the pre-processing before setting any value is done via
        this api.
        """
        return value

    def post_set(self, obj, value):
        """All the post-processing after setting any value is done via
        this api.
        """
        return value

    def pre_get(self, obj, value):
        """All the pre-processing before getting any value is done via
        this api.
        """
        return value

    def post_get(self, obj, value):
        """All the pre-processing after getting any value is done via
        this api.
        """
        return value

    def pre_delete(self, obj, value):
        """All the pre-processing before deleting any value is done via
        this api.
        """
        return value

    def post_delete(self, obj, value):
        """All the pre-processing after deleting any value is done via
        this api.
        """
        return value

    def _set_name(self, owner, name, logger):
        if self.errors is None:
            self.errors = []

        # set name
        try:
            if self.name is None:
                self.name = name
                if logger:
                    logger.info(f"assigned: {owner.__name__}.{name}")

            # # if name is set ensure names match.
            elif name != self.name and self.annotation == owner.__annotations__.get(name, None):
                raise AttributeError(
                    f"{self.name} != {name}, attribute names did not match"
                )

        except (RuntimeError, errors.SetPropertyError, Exception) as name_err:
            if logger:
                logger.error(name_err)
            self.errors.append(name_err)
            raise name_err

    def _may_set_or_ensure_annotation_match(self, owner, name, logger):
        if self.errors is None:
            self.errors = []

        # set annotations
        try:
            # if owner has __annotations__ set, ensure annotations match
            # else assign annotations if any available

            if "__annotations__" not in owner.__dict__:
                owner.__annotations__ = dict()
            if name in owner.__annotations__:
                if self.annotation is None:
                    self.annotation = owner.__annotations__[name]
                else:
                    self_annotation = self.annotation
                    owner_annotation = owner.__annotations__[name]
                    if not issubclassx(owner_annotation, self_annotation):
                        owner_annotation_name = (
                            owner_annotation.__name__
                            if hasattr(owner_annotation, "__name__")
                            else str(owner_annotation)
                        )
                        self_annotation_name = (
                            self_annotation.__name__
                            if hasattr(self_annotation, "__name__")
                            else str(self_annotation)
                        )
                        raise TypeError(
                            f"{owner.__name__}.{self.name}: {owner_annotation_name}"
                            f" annotation did not match {type(self).__qualname__}: "
                            f"{self_annotation_name}"
                        )
                    if owner_annotation != self.annotation:
                        self.annotation = owner_annotation
            else:
                if self.annotation:
                    owner.__annotations__[name] = self.annotation
            # https://github.com/python/cpython/blob/1b293b60067f6f4a95984d064ce0f6b6d34c1216/Lib/typing.py#L1787-L1811
            # self.__supertype__ = owner.__annotations__[name]
            # self.__name__ = name
        except (RuntimeError, errors.SetPropertyError, Exception) as annotation_err:
            if logger:
                logger.error(annotation_err)
            self.errors.append(annotation_err)
            raise annotation_err

    def _set_docs(self, owner, name, logger):
        owner_annotation = owner.__annotations__[name]
        if self.errors is None:
            self.errors = []

        # set or extend docs for the class
        try:
            owner.__doc__ = owner.__doc__ or f"Class: {owner.__name__}"

            owner.__doc__ += (
                (
                    f"\n\t:param {owner_annotation} {str(self)}: {self.doc}"
                    if self.annotation
                    else f"\n\t:param {str(self)}: {self.doc}"
                )
                if self.doc is not None
                else (
                    f"\n\t:param {owner_annotation} {str(self)}:"
                    if self.annotation
                    else f"\n\t:param {str(self)}:"
                )
            )
        except (Exception,) as doc_err:
            if logger:
                logger.error(doc_err)
            self.errors.append(doc_err)
            raise doc_err

    def __set_name__(self, owner, name):
        logger = (
            self.get_logger(owner.__name__, name)
            if name is not None and self.logger is not False
            else None
        )
        if logger:
            _name = (
                f"{type(self).__name__}({self.name})"
                if self.name is not None
                else f"{type(self).__name__}"
            )
            logger.info(f"{_name}: assigning as property : {owner.__name__}.{name}")

        if self.errors is None:
            self.errors = []

        self._set_name(owner=owner, name=name, logger=logger)
        self._may_set_or_ensure_annotation_match(owner=owner, name=name, logger=logger)
        self._set_docs(owner=owner, name=name, logger=logger)

    def __set__(self, obj, value):
        """sets descriptor field"""
        class_name = type(obj).__name__
        attr_name = self.name

        if logger := self.logger:
            logger.info(f"setting: {class_name}.{attr_name}")
        
        try:
            value = value or (self.default \
                if not callable(self.default) else self.default()) \
                if self.default is not None else value
            value = self.pre_set(obj, value)
            obj.__dict__[self.name] = value
            if logger:
                logger.info(f"set: {class_name}.{attr_name}")
            self.post_set(obj, value)
        except (errors.SetPropertyError, Exception) as spe:
            if logger:
                logger.error(spe)
            self.errors.append(spe)
            if self.debug:
                raise spe

    def __get__(self, obj, obj_type=None):
        """gets descriptor field"""
        if obj is None:
            return #self
        
        class_name = type(obj).__name__
        attr_name = self.name
        if logger := self.logger:
            logger.info(f"getting: {class_name}.{attr_name}")
        error = False
        try:
            self.pre_get(obj, self.name)
            return obj.__dict__[self.name]
        except (errors.GetPropertyError, Exception) as gpe:
            if logger:
                logger.error(gpe)
            error = True
            self.errors.append(gpe)
            if self.debug:
                raise gpe
        finally:
            if logger and not error:
                logger.info(f"got: {class_name}.{attr_name}")
            try:
                self.post_get(obj, self.name)
            except (Exception,) as post_get_err:
                if logger:
                    logger.error(post_get_err)
                self.errors.append(post_get_err)
                if self.debug:
                    raise post_get_err

    def __delete__(self, obj):
        """deletes descriptor"""
        class_name = type(obj).__name__
        attr_name = self.name
        if logger := self.logger:
            logger.info(f"deleting: {class_name}.{attr_name}")
        error = False
        try:
            self.pre_delete(obj, self.name)
            del obj.__dict__[self.name]
            self.post_delete(obj, self.name)

        except (errors.DeletePropertyError, Exception) as dpe:
            if logger:
                logger.error(dpe)
            error = True
            self.errors.append(dpe)
            if self.debug:
                raise dpe
        finally:
            if logger and not error:
                logger.info(f"deleted: {class_name}.{attr_name}")

    def __str__(self):
        return str(self.name)

    def dict(self):
        if self._dict is None:
            self._dict = dict(zip(map(
                lambda field: field.name, fields(self)),
                astuple(self)), **self.kwargs)
        return self._dict

    def __repr__(self):
        return pformat(self._dict)

# if __name__ == "__main__":
#     from dataclasses import dataclass
#
#     name_field = Property(annotation=str)
#     salary_field = Property(annotation=int, log_levels=["INFO", "DEBUG"])
#
#     @dataclass
#     class Person(object):
#         name: str = name_field
#         salary: int = salary_field
#
#     t = Person()
#     t.name = "some shitty name"
#     t.salary = 2222
#     print(t.name)
#     print(t.__doc__)
#     print(Property(name="2", doc="helping") == Property("2"))
#     import pprint
# pprint.pprint(salary_field.as_dict())



# Copyright (c) 2022 Valio
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import abc
import datetime
import pathlib
from dataclasses import dataclass, make_dataclass
from typing import Any, Type

from valio.regexer.regexps import IfFollowedBy, Pattern, PatternType
from valio.regexer.relib import DigitGroups, WordGroups
from valio.validator import validators

REGEX_MAP = {str: WordGroups, int: DigitGroups}

__all__ = [
    "BytesSchema",
    "BooleanSchema",
    "IntegerSchema",
    "FloatSchema",
    "CharSchema",
    "FileSchema",
    "DateSchema",
    "EmailSchema",
    "PhoneNumberSchema",
    "PaymentCardSchema",
    "SchemaMixin",
]


class SchemaValidator(object):
    name = validators.StringValidator(logger=False)
    validator = Type[validators.Validator]
    annotation = Any

    def __set_name__(self, owner, name):
        self.name = name
        self.validator = make_dataclass(
            "validator", [(self.name, self.annotation, self.validator(**self.kwargs))]
        )

    def __init__(self, **kwargs):
        self.name = kwargs.pop("name", None) or self.name or type(self).__name__
        self.kwargs = kwargs


class SchemaBase(SchemaValidator):

    def _is_valid(self, *args):
        for v in args:
            self.validator(v)

    @property
    def args(self):
        return self._args

    @args.setter
    def args(self, args):
        self._is_valid(*args)
        self._args = args

    def __init__(self, *args, **kwargs):
        self.kwargs = kwargs
        regex_map = (
            {self.annotation: REGEX_MAP[self.annotation]}
            if self.annotation in REGEX_MAP
            else {
                self.annotation: {
                    v: REGEX_MAP[self.__class__.__dict__[v].annotation]
                    for v in self.__class__.__dict__
                    if isinstance(self.__class__.__dict__[v], SchemaBase)
                       and self.__class__.__dict__[v].annotation in REGEX_MAP
                }
            }
        )

        self.regex = (
            REGEX_MAP[self.annotation](
                count_min=self.kwargs.get("min_length", None),
                count_max=self.kwargs.get("max_length", None),
                name=str(self),
            ).named_capturing_group
            if self.annotation in REGEX_MAP
            else regex_map[self.annotation]
        )
        if isinstance(self.regex, dict):
            self.regex.update({k: getattr(self, k).regex for k in self.regex})

        pattern_field = self.kwargs.get("pattern", None)
        if pattern_field is not None and not isinstance(
                pattern_field, (str, PatternType)
        ):
            raise TypeError(
                f"expected 'pattern' of {str.__name__} or {Pattern.__name__} types,"
                f" got {type(pattern_field).__name__} type instead"
            )
        if not isinstance(self.regex, dict):
            if any(
                    {
                        v: getattr(pattern_field, v)
                        for v in self.__class__.__dict__
                        if (
                            isinstance(self.__class__.__dict__[v], SchemaBase)
                            and getattr(pattern_field, v, None) is not None
                    )
                    }
            ):
                raise AttributeError(
                    f"{self.regex.__name__} and {pattern_field.__name__} fields did not match"
                )

        regex = (
            (
                self.regex
                if pattern_field is None
                else IfFollowedBy(Pattern(pattern=pattern_field)) & self.regex
                if not isinstance(pattern_field, PatternType)
                else IfFollowedBy(pattern=pattern_field) & self.regex
            )
            if not isinstance(self.regex, dict)
            else {
                k: self.regex[k] & getattr(pattern_field, k)
                if getattr(pattern_field, k, None) is not None
                else self.regex[k]
                for k in self.regex
            }
        )

        self.regex = regex

        super(SchemaBase, self).__init__(**self.kwargs)
        if hasattr(self, "schema"):
            self._validate = getattr(self, "schema")

        self.args = args

    def __str__(self):
        return self.kwargs.get("name", None) or self.name or str(type(self).__name__)

    def __repr__(self):
        return f"{self.__str__()}({self.kwargs})"

    def __iter__(self):
        return iter(self.args)

    def __next__(self):
        return next(self)

    def __call__(self, *value):
        self.args = value
        return value


class BytesSchema(SchemaBase):
    annotation = bytes
    validator = validators.BytesValidator


class BooleanSchema(SchemaBase):
    annotation = bool
    validator = validators.BooleanValidator


class NumberSchema(SchemaBase):
    pass


class IntegerSchema(NumberSchema):
    annotation = int
    validator = validators.IntegerValidator


class PositiveIntegerSchema(IntegerSchema):
    def __init__(self, *args, **kwargs):
        super(PositiveIntegerSchema, self).__init__(*args, min_value=0, **kwargs)


class NegativeIntegerSchema(IntegerSchema):
    def __init__(self, *args, **kwargs):
        super(NegativeIntegerSchema, self).__init__(*args, max_value=0, **kwargs)


class FloatSchema(NumberSchema):
    annotation = float
    validator = validators.FloatValidator


class PositiveFloatSchema(FloatSchema):
    def __init__(self, *args, **kwargs):
        super(PositiveFloatSchema, self).__init__(*args, min_value=0.0, **kwargs)


class NegativeFloatSchema(FloatSchema):
    def __init__(self, *args, **kwargs):
        super(NegativeFloatSchema, self).__init__(*args, max_value=0.0, **kwargs)


class CharSchema(SchemaBase):
    annotation = str
    validator = validators.StringValidator


class DateSchema(SchemaBase):
    annotation = datetime.datetime
    validator = validators.DateValidator


class FileSchema(SchemaBase):
    annotation = pathlib.Path
    validators = validators.PathValidator


class EmailSchema(SchemaBase):
    annotation = str
    validator = validators.EmailIDValidator


class PaymentCardSchema(SchemaBase):
    annotation = str
    validator = validators.PaymentCardValidator


class PhoneNumberSchema(SchemaBase):
    annotation = str
    validator = validators.PhoneNumberValidator


class SchemaMixin(SchemaBase, abc.ABC):
    def __init__(self, *args, **kwargs):
        super(SchemaMixin, self).__init__(*args, **kwargs)

    @abc.abstractmethod
    def schema(self, *args, **kwargs):
        pass


if __name__ == "__main__":
    from secrets import compare_digest


    @dataclass
    class PasswordField(object):
        password: str
        confirm_password: str
        schema: "PasswordSchema" = None

        def __str__(self):
            raise PermissionError("permission denied")

        def __post_init__(self):
            self.schema(self)


    class PasswordSchema(SchemaMixin):
        password = CharSchema(
            min_length=5,
            max_length=32,
            pattern=Pattern(r"\d+\w+"),
            not_in_choice=["", '', "", "*"],
            logger=True,
            debug=True,
        )
        confirm_password = CharSchema(
            min_length=5,
            max_length=32,
            pattern=Pattern(r"\d+\w+"),
            logger=True,
            debug=True
        )
        annotation = PasswordField

        def schema(self, password: PasswordField):
            self.password(password.password)
            self.confirm_password(password.confirm_password)
            if not compare_digest(password.password, password.confirm_password):
                raise ValueError(
                    f"{self.password.name} and {self.confirm_password.name} did not match"
                )


    PasswordField.schema = PasswordSchema()
    password_ = PasswordField("9989z", "9989z")
    print(password_.schema)



# Copyright (c) 2022 Valio
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT



from valio.regexer.regexps import Pattern, SetOf

# Specials (32-47)


space = Pattern(pattern=r"\s")
exclamation_point = Pattern(pattern=r"!")
quotation_mark = Pattern(pattern='"')
pound_sign = Pattern(pattern=r"#")
dollar_sign = Pattern(pattern=r"$")
percent_sign = Pattern(pattern=r"%")
ampersand = Pattern(pattern=r"&")
apostrophe = Pattern(pattern=r"'")
left_parenthesis = Pattern(pattern=fr"(")
right_parenthesis = Pattern(pattern=fr")")
asterisk = Pattern(pattern=r"*")
plus_sign = Pattern(pattern=r"+")
comma = cedilla = Pattern(pattern=r",")
hyphen = minus_sign = Pattern(pattern=r"-")
period = decimal_point = Pattern(pattern=r".")
slant = solidus = Pattern(pattern=r"/")

special_chars = (
        exclamation_point
        & quotation_mark
        & pound_sign
        & dollar_sign
        & percent_sign
        & ampersand
        & apostrophe
        & left_parenthesis
        & right_parenthesis
        & asterisk
        & plus_sign
        & comma
        & hyphen
        & period
        & slant
)

special_chars_set = SetOf(special_chars)



# Copyright (c) 2022 Valio
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT



class SetPropertyError(AttributeError):
    pass


class GetPropertyError(AttributeError):
    pass


class DeletePropertyError(AttributeError):
    pass


class SetAttributeError(AttributeError):
    pass


class GetAttributeError(AttributeError):
    pass


class DeleteAttributeError(AttributeError):
    pass


class DustBaseException(Exception):
    """Base Dust exception"""

    def __init__(self, error_message=None):
        super(DustBaseException, self).__init__(error_message)


class DustError(DustBaseException):
    """helper for validation parsing"""

    def __init__(self, error_message=None, error_param=None, log_message=None):
        # error_message api provides custom error message for value error
        self._error_message = None

        # error_param api provides error parameters
        self._error_param = None

        # log_message api provides a way to log messages
        self._log_message = None

        # input errors
        self._error_message_error = TypeError(
            f"expect 'message' to be '{str.__name__}' type, "
            f"got '{type(error_message).__name__}' type instead"
        )

        self._error_param_error = TypeError(
            f"expect 'param' parameter to be '{tuple.__name__}' type, "
            f"got '{type(error_param)}' type instead"
        )

        self._log_message_error = TypeError(
            f"expect 'log' to be '{str.__name__}' type, "
            f"got '{type(error_message).__name__}' type instead"
        )
        self.message = error_message
        self.param = error_param
        self.log = log_message

        super(DustError, self).__init__(self.message)

    @property
    def message(self):
        if all([self._error_message, self._error_param]):
            return self._error_message % self._error_param
        return self._error_message

    @message.setter
    def message(self, error_message=None):
        if error_message is not None and self._error_message is None:
            if not isinstance(error_message, str):
                raise self._error_message_error
            self._error_message = error_message

    @property
    def param(self):
        return self._error_param

    @param.setter
    def param(self, error_param):
        if error_param is not None and self._error_param is None:
            if not isinstance(error_param, (tuple, list)):
                raise self._error_param_error
            self._error_param = error_param

    @property
    def log(self):
        return self._log_message

    @log.setter
    def log(self, log_message):
        if log_message is not None and self._log_message is None:
            if not isinstance(log_message, str):
                raise self._log_message_error
        self._log_message = log_message



from valito.regexer.regexps import Pattern, NonCapturingGroup, SetOf, CapturingGroup, WordBoundary
from valito.regexer.relib.special_chars import pound_sign

space = Pattern(r"\s", count_min=0)
zx = Pattern(r"0x")
_hex_start = NonCapturingGroup((pound_sign | zx), greedy=False)
O_9 = Pattern(r"0-9")
a_f = Pattern(r"a-f")
O_9a_f_single_set = SetOf(O_9 & a_f)
O_9a_f_double_set = SetOf(O_9 & a_f, count=2)
_rgb_ss = CapturingGroup(O_9a_f_single_set)
_rgb_ss_opn = CapturingGroup(O_9a_f_single_set, greedy=False)
_rgb_ds = CapturingGroup(O_9a_f_double_set)
_rgb_ds_opn = CapturingGroup(O_9a_f_double_set, greedy=False)

####################################################################################
r_hex_short = space & _hex_start & _rgb_ss & _rgb_ss & _rgb_ss & _rgb_ss_opn & space
r_hex_long = space & _hex_start & _rgb_ds & _rgb_ds & _rgb_ds & _rgb_ds_opn & space
####################################################################################


O_1 = SetOf(Pattern(r"0-1"), count=1, greedy=False)
_2 = SetOf(Pattern(r"2"), count=1, greedy=False)
O_99 = SetOf(Pattern(r"0-9"), count=2, greedy=False)
O_55 = SetOf(Pattern(r"0-5"), count=2, greedy=False)
decimal_optional = NonCapturingGroup(Pattern(r"\.") & Pattern(r"\d", count_min=1), greedy=False)
O_255 = WordBoundary(((O_1 & O_99) & decimal_optional) | ((_2 & O_55) & decimal_optional))
_r_255 = fr'(\d{1,3}{decimal_optional})'
_r_comma = space & Pattern(r",") & space
r_rgb = fr'{space}rgb\({space}{_r_255}{_r_comma}{_r_255}{_r_comma}{_r_255}\){space}'
_r_alpha = fr'(\d{decimal_optional}|\.\d+|\d{1,2}%)'
r_rgba = fr'{space}rgba\({space}{_r_255}{_r_comma}{_r_255}{_r_comma}{_r_255}{_r_comma}{_r_alpha}{space}\){space}'
_r_h = fr'(-?\d+{decimal_optional}|-?\.\d+)(deg|rad|turn)?'
_r_sl = fr'(\d{1,3}{decimal_optional})%'
r_hsl = fr'{space}hsl\({space}{_r_h}{_r_comma}{_r_sl}{_r_comma}{_r_sl}{space}\){space}'
r_hsla = fr'{space}hsl\({space}{_r_h}{_r_comma}{_r_sl}{_r_comma}{_r_sl}{_r_comma}{_r_alpha}{space}\){space}'

# colors where the two hex characters are the same, if all colors match this the short version of hex colors can be used
repeat_colors = {int(c * 2, 16) for c in '0123456789abcdef'}



# Copyright (c) 2022 Valio
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT



import pytest
from valio.regexer import regexps, relib
from valio.regexer.relib import patterns


def assert_pattern_has(
        quantifier: str,
        if_greedy: bool,
        with_count_min: int = None,
        with_count_max: int = None,
):
    O_9 = fr"0-9"  # NOQA
    a_zA_Z = fr"a-zA-Z"  # NOQA
    w = fr"\w"  # NOQA
    d = fr"\d"  # NOQA
    s = fr"\s"  # NOQA
    W = fr"\W"  # NOQA
    D = fr"\D"  # NOQA
    S = fr"\S"  # NOQA
    p_0_9 = regexps.Pattern(
        O_9, count_min=with_count_min, count_max=with_count_max, greedy=if_greedy
    ).pattern
    p_a_zA_Z = patterns.Pattern(  # NOQA
        a_zA_Z, count_min=with_count_min, count_max=with_count_max, greedy=if_greedy
    ).pattern
    p_w = regexps.Pattern(
        w, count_min=with_count_min, count_max=with_count_max, greedy=if_greedy
    ).pattern
    p_d = regexps.Pattern(
        d, count_min=with_count_min, count_max=with_count_max, greedy=if_greedy
    ).pattern
    p_s = regexps.Pattern(
        s, count_min=with_count_min, count_max=with_count_max, greedy=if_greedy
    ).pattern
    p_W = regexps.Pattern(  # NOQA
        W, count_min=with_count_min, count_max=with_count_max, greedy=if_greedy
    ).pattern
    p_D = regexps.Pattern(  # NOQA
        D, count_min=with_count_min, count_max=with_count_max, greedy=if_greedy
    ).pattern
    p_S = regexps.Pattern(  # NOQA
        S, count_min=with_count_min, count_max=with_count_max, greedy=if_greedy
    ).pattern

    assert p_0_9 == f"{O_9}{quantifier}", f"{O_9}{quantifier} != {p_0_9} error"
    assert (
            p_a_zA_Z == f"{a_zA_Z}{quantifier}"
    ), f"{a_zA_Z}{quantifier} != {p_a_zA_Z} error"
    assert p_w == f"{w}{quantifier}", f"{w}{quantifier} != {p_w} error"
    assert p_d == f"{d}{quantifier}", f"{d}{quantifier} != {p_d} error"
    assert p_s == f"{s}{quantifier}", f"{s}{quantifier} != {p_s} error"
    assert p_W == f"{W}{quantifier}", f"{W}{quantifier} != {p_W} error"
    assert p_D == f"{D}{quantifier}", f"{D}{quantifier} != {p_D} error"
    assert p_S == f"{S}{quantifier}", f"{S}{quantifier} != {p_S} error"


def test_pattern_for_set():
    O_9 = fr"0-9"  # NOQA
    a_zA_Z = fr"a-zA-Z"  # NOQA
    w = fr"\w"  # NOQA
    d = fr"\d"  # NOQA
    s = fr"\s"  # NOQA
    W = fr"\W"  # NOQA
    D = fr"\D"  # NOQA
    S = fr"\S"  # NOQA
    assert regexps.Pattern(O_9).raw_pattern == O_9, f"{O_9} error"
    assert regexps.Pattern(a_zA_Z).raw_pattern == a_zA_Z, f"{a_zA_Z} error"
    assert regexps.Pattern(w).raw_pattern == w, f"{w} error"
    assert regexps.Pattern(d).raw_pattern == d, f"{d} error"
    assert regexps.Pattern(s).raw_pattern == s, f"{s} error"
    assert regexps.Pattern(W).raw_pattern == W, f"{W} error"
    assert regexps.Pattern(D).raw_pattern == D, f"{D} error"
    assert regexps.Pattern(S).raw_pattern == S, f"{S} error"


def test_set_of():
    O_9 = fr"0-9"  # NOQA
    a_zA_Z = fr"a-zA-Z"  # NOQA
    w = fr"\w"  # NOQA
    d = fr"\d"  # NOQA
    s = fr"\s"  # NOQA
    W = fr"\W"  # NOQA
    D = fr"\D"  # NOQA
    S = fr"\S"  # NOQA
    assert regexps.SetOf(regexps.Pattern(O_9)).pattern == f"[{O_9}]", f"[{O_9}] error"
    assert (
            regexps.SetOf(regexps.Pattern(a_zA_Z)).pattern == fr"[{a_zA_Z}]"
    ), f"[{a_zA_Z}] error"
    assert regexps.SetOf(regexps.Pattern(w)).pattern == fr"[{w}]", f"{w} error"
    assert regexps.SetOf(regexps.Pattern(d)).pattern == fr"[{d}]", f"{d} error"
    assert regexps.SetOf(regexps.Pattern(s)).pattern == fr"[{s}]", f"{s} error"
    assert regexps.SetOf(regexps.Pattern(W)).pattern == fr"[{W}]", f"{W} error"
    assert regexps.SetOf(regexps.Pattern(D)).pattern == fr"[{D}]", f"{D} error"
    assert regexps.SetOf(regexps.Pattern(S)).pattern == fr"[{S}]", f"{S} error"


def test_pattern():
    assert_pattern_has(quantifier=fr"?", if_greedy=False)
    assert_pattern_has(quantifier=fr"", if_greedy=True)
    assert_pattern_has(quantifier=fr"*?", with_count_min=0, if_greedy=False)
    assert_pattern_has(quantifier=fr"*", with_count_min=0, if_greedy=True)
    assert_pattern_has(quantifier=fr"+?", with_count_min=1, if_greedy=False)
    assert_pattern_has(quantifier=fr"+", with_count_min=1, if_greedy=True)
    assert_pattern_has(quantifier=fr"??", with_count_max=1, if_greedy=False)
    assert_pattern_has(quantifier=fr"?", with_count_max=1, if_greedy=True)
    assert_pattern_has(
        quantifier=fr"??", with_count_min=0, with_count_max=1, if_greedy=False
    )
    assert_pattern_has(
        quantifier=fr"?", with_count_min=0, with_count_max=1, if_greedy=True
    )
    assert_pattern_has(
        quantifier=fr"{{{0},{9}}}", with_count_min=0, with_count_max=9, if_greedy=True
    )
    assert_pattern_has(
        quantifier=f"{{{1}}}?", with_count_min=1, with_count_max=1, if_greedy=False
    )
    assert_pattern_has(
        quantifier=f"{{{1}}}", with_count_min=1, with_count_max=1, if_greedy=True
    )


def test_preceded_by():
    pattern_a = regexps.IfPrecededBy(
        regexps.All(
            regexps.EndsWith(
                regexps.SetOf((patterns.WordGroups() | patterns.DigitGroups()))
            ),
            patterns.WhiteSpace(count_max=2),
            regexps.SetOf(regexps.Pattern(fr"\w", count_min=1))
            & regexps.SetOf(regexps.Pattern(fr"\d")),
        )
    ).pattern

    preceding_special_chars = patterns.IfPreceding(
        relib.special_chars_set, name="special_chars_set"
    )

    pattern_b = regexps.IfGroupNameMatched(
        preceding_special_chars.named_capturing_group,
        regexps.SetOf(regexps.Pattern(fr"\w", count_max=1, greedy=False)),
        patterns.a_zA_Z,
    ).pattern

    pattern_c = regexps.SetOf(patterns.Word(count_min=2)) & regexps.SetOf(
        patterns.Digit(count_min=2)
    ) | ~regexps.SetOf(patterns.Word(count_min=1)) & regexps.NonCapturingGroup(
        # name="Year",
        pattern=regexps.WordBoundary(
            regexps.StartsWith(
                regexps.SetOf(patterns.Digit(count_min=2, count_max=4, greedy=False))
            )
            & regexps.Pattern(pattern=fr":", count_min=1, count_max=1)
            & regexps.EndsWith(~regexps.SetOf(patterns.Word(count_min=4)))
        ),
    )
    assert (
            pattern_a == fr"(?<=[\w|\d]$\s{{{0},{2}}}[\w]+[\d])"
    ), fr"{pattern_a} != (?<=[\w|\d]$\s{{{0},{2}}}[\w]+[\d])"
    assert (
            preceding_special_chars.named_capturing_group.pattern
            == r"(?P<special_chars_set>(?<=[!#$%&'*+/=?^_`{|}~-]))"
    ), fr"{preceding_special_chars.named_capturing_group} != (?P<special_chars_set>(?<=[!#$%&'*+/=?^_`{{|}}~-]))"
    assert (
            pattern_b == r"(?(special_chars_set)[\w]??|[a-zA-Z])"
    ), fr"{pattern_b} != (?(special_chars_set)[\w]??|[a-zA-Z])"


def main_test():
    test_pattern()
    test_set_of()
    test_pattern_for_set()
    # test_preceded_by()


if __name__ == "__main__":
    pytest.main(main_test())



# Copyright (c) 2022 Valio
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MITs

from .aadhaarcard import *
from .control_chars import *
from .dates import *
from .emails import *
from .ipaddresses import *
from .pancard import *
from .patterns import *
from .paymentcards import *
from .special_chars import *



# Copyright (c) 2022 Valio
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from datetime import datetime
from re import IGNORECASE

import pyparsing as pp
from valio.regexer.regexps import (NamedCapturingGroup, NonCapturingGroup,
                                   Pattern, SetOf, WordBoundary)

__all__ = [
    "months",
    "years",
    "jan",
    "feb",
    "march",
    "april",
    "may",
    "june",
    "july",
    "august",
    "september",
    "october",
    "november",
    "december",
    "day_numbers",
    "months_names",
    "months_numbers",
    "dates",
    "times",
    "eu_date",
    "ind_date",
    "get_date"
]

#####################################################################################
#               another better implementation with leap year filtering              #
#                          https://stackoverflow.com/a/10309472                     #
#                   with valid days and months formats.                             #
#               https://www.regular-expressions.info/branchreset.html               #
#####################################################################################

#####################################################################################
space = Pattern(r"\s", count_min=0, greedy=False, alias=" ")
optional_zero = Pattern(r"0", count_max=1, greedy=False)
_monthsToNum = {
    "jan": 1,
    "feb": 2,
    "mar": 3,
    "apr": 4,
    "may": 5,
    "jun": 6,
    "jul": 7,
    "aug": 8,
    "sep": 9,
    "oct": 10,
    "nov": 11,
    "dec": 12
}
###################################################################################

# days
first_nine_days = optional_zero & SetOf(Pattern(r"1-9"), count=1)
tens_and_twenties = SetOf(Pattern(r"1-2"), count=1) & SetOf(Pattern(r"0-9"), count=1)
thirties = Pattern("3") & SetOf(Pattern(r"01"), count=1)
day_numbers = WordBoundary(
    space
    & Pattern(NonCapturingGroup(first_nine_days | tens_and_twenties | thirties).pattern, alias="DD")
    & space
)
rd = WordBoundary(space & NonCapturingGroup(Pattern(r"rd", alias="rd"), count_max=1, greedy=False) & space)
th = WordBoundary(space & NonCapturingGroup(Pattern(r"th", alias="th"), count_max=1, greedy=False) & space)
st = WordBoundary(space & NonCapturingGroup(Pattern(r"st", alias="st"), count_max=1, greedy=False) & space)
nd = WordBoundary(space & NonCapturingGroup(Pattern(r"nd", alias="nd"), count_max=1, greedy=False) & space)
day_postfix = rd | th | st | nd

# months
first_nine_months = optional_zero & SetOf(Pattern(r"1-9"), count=1)
tenth_to_twelth_months = Pattern(r"1", count=1) & SetOf(Pattern(r"0-2"), count=1)
months_numbers = Pattern((first_nine_months | tenth_to_twelth_months).pattern, alias="MM")

# january month
jan = WordBoundary(
    (Pattern(r"J", count=1) | Pattern(r"j", count=1))
    & Pattern(r"an")
    & NonCapturingGroup(Pattern(r"uary"), count_max=1, greedy=False)
)

# february month
feb = WordBoundary(
    (Pattern(r"F", count=1) | Pattern(r"f", count=1))
    & Pattern(r"eb")
    & NonCapturingGroup(Pattern(r"ruary"), count_max=1, greedy=False)
)

# march month
march = WordBoundary(
    (Pattern(r"M", count=1) | Pattern(r"m", count=1)) & Pattern(r"ar")
    & NonCapturingGroup(Pattern(r"ch"), count_max=1, greedy=False)
)

# april month
april = WordBoundary(
    (Pattern(r"A") | Pattern(r"a")) & NonCapturingGroup(Pattern(r"pr"), count=1)
    & NonCapturingGroup(Pattern(r"il"), count_max=1, greedy=False)
)

# may month
may = WordBoundary(
    (Pattern(r"M") | Pattern(r"m")) & NonCapturingGroup(Pattern(r"ay"), count=1)
)

# june month
june = WordBoundary(
    (Pattern(r"J") | Pattern(r"j"))
    & NonCapturingGroup(Pattern(r"un"), count=1)
    & NonCapturingGroup(Pattern(r"e"), count_max=1, greedy=False)
)

# july month
july = WordBoundary(
    (Pattern(r"J") | Pattern(r"j"))
    & NonCapturingGroup(Pattern(r"ul"))
    & NonCapturingGroup(Pattern(r"y"), count_max=1, greedy=False)
)

# august month
august = WordBoundary(
    (Pattern(r"A") | Pattern(r"a"))
    & NonCapturingGroup(Pattern(r"ug"))
    & NonCapturingGroup(Pattern(r"ust"), count_max=1, greedy=False)
)

# september month
september = WordBoundary(
    (Pattern(r"S") | Pattern(r"s"))
    & NonCapturingGroup(Pattern(r"ep"), count=1)
    & (NonCapturingGroup(Pattern(r"t"), count_max=1, greedy=False)
       | NonCapturingGroup(Pattern(r"tember"), count_max=1, greedy=False)
       )
)

# october month
october = WordBoundary(
    (Pattern(r"O") | Pattern(r"o"))
    & NonCapturingGroup(Pattern(r"ct"), count=1)
    & NonCapturingGroup(Pattern(r"ober"), count_max=1, greedy=False)
)

# november month
november = WordBoundary(
    (Pattern(r"N") | Pattern(r"n"))
    & NonCapturingGroup(Pattern(r"ov"), count=1)
    & NonCapturingGroup(Pattern(r"ember"), count_max=1, greedy=False)
)

# december month
december = WordBoundary(
    (Pattern(r"D") | Pattern(r"d"))
    & NonCapturingGroup(Pattern(r"ec"), count=1)
    & NonCapturingGroup(Pattern(r"ember"), count_max=1, greedy=False)
)

months_names = jan | feb | march | april | may | june | july | august | september \
               | october | november | december
months = WordBoundary(space & (months_numbers | months_names) & space)
months_names_only = WordBoundary(space & months_names & space)
years = WordBoundary(space & Pattern(r"\d", count=4, alias="YYYY") & space)

# delimiters
hyphen = NonCapturingGroup(Pattern(r"-", alias="-"), count=1)
colon = NonCapturingGroup(Pattern(r":", alias=":"), count=1)
backslash = NonCapturingGroup(Pattern(r"\/", alias="/"), count=1)
dot = NonCapturingGroup(Pattern(r"\.", alias="."), count=1)
comma = NonCapturingGroup(Pattern(r",", alias=","), count=1, greedy=False)

# European Date Format
eu_date_with_hyphen = years & hyphen & months & hyphen & ((day_numbers & day_postfix) | day_numbers)
eu_date_with_colon = years & colon & months & colon & ((day_numbers & day_postfix) | day_numbers)
eu_date_with_dot = years & dot & months & dot & ((day_numbers & day_postfix) | day_numbers)
eu_date_with_backslash = years & backslash & months & backslash & ((day_numbers & day_postfix) | day_numbers)
eu_date_with_comma = years & ((space & comma & space) | comma | space) & months_names_only & (
        (space & comma & space) | comma | space) & ((day_numbers & day_postfix) | day_numbers)
eu_date_with_space = years & space & months_names_only & ((day_numbers & day_postfix) | day_numbers)
eu_date = NamedCapturingGroup(name="eu_fmt_date",
                              pattern=(eu_date_with_colon
                                       | eu_date_with_hyphen
                                       | eu_date_with_backslash
                                       | eu_date_with_dot
                                       | eu_date_with_comma
                                       | eu_date_with_space
                                       )
                              )

# Indian Date format
ind_date_with_hyphen = ((day_numbers & day_postfix) | day_numbers) & hyphen & months & hyphen & years
ind_date_with_colon = ((day_numbers & day_postfix) | day_numbers) & colon & months & colon & years
ind_date_with_dot = ((day_numbers & day_postfix) | day_numbers) & dot & months & dot & years
ind_date_with_backslash = ((day_numbers & day_postfix) | day_numbers) & backslash & months & backslash & years
ind_date_with_comma = ((day_numbers & day_postfix) | day_numbers) & (
        (space & comma & space) | comma | space) & months_names_only & (
                              (space & comma & space) | comma | space) & years
ind_date_with_space = ((day_numbers & day_postfix) | day_numbers) & months_names_only & years
ind_date = NamedCapturingGroup(name="ind_fmt_date",
                               pattern=(ind_date_with_hyphen
                                        | ind_date_with_colon
                                        | ind_date_with_backslash
                                        | ind_date_with_dot
                                        | ind_date_with_comma
                                        | ind_date_with_space
                                        )
                               )

date_Date_dated_Dated = (Pattern("d", count=1) | Pattern("D", count=1)) \
                        & Pattern("ate") & Pattern("d", count_min=0, greedy=False)

dob = (
        (Pattern(r"d", count=1) | Pattern(r"D", count=1))
        & (dot | space) & (Pattern(r"o", count=1) | Pattern(r"O", count=1))
        & (dot | space) & (Pattern(r"b", count=1, greedy=False) | Pattern(r"B", count=1, greedy=False))
)

# times pattern
sextile_unit = SetOf(Pattern(r"0-9"), count=1)
sextile_tens = SetOf(Pattern(r"0-5"), count=1)
O_59 = WordBoundary(sextile_unit | (sextile_tens & sextile_unit))
minutes = space & NonCapturingGroup(pattern=O_59) & space
seconds = space & NonCapturingGroup(pattern=O_59) & space
hours = space & SetOf(Pattern(r"0-1"), count=1) & SetOf(Pattern(r"0-9"), count=1) \
        | Pattern(r"2", count=1) & SetOf(Pattern(r"0-3"), count=1) & space
am = WordBoundary(NonCapturingGroup(pattern=Pattern(r"am") | Pattern(r"AM"), greedy=False))
pm = WordBoundary(NonCapturingGroup(pattern=Pattern(r"pm") | Pattern(r"PM"), greedy=False))
am_or_pm = space & NamedCapturingGroup(name="am_or_pm", pattern=(am | pm)) & space
times = NamedCapturingGroup(name="times",
                            pattern=WordBoundary((
                                                         (hours & colon & minutes & colon & seconds)
                                                         | (hours & colon & minutes)
                                                         | hours) & am_or_pm
                                                 )

                            , greedy=False)

date_prefix = NamedCapturingGroup(
    name="date_prefix",
    pattern=WordBoundary(space & (date_Date_dated_Dated | dob)
                         & space & (colon | hyphen | space | colon & hyphen)
                         & space),
    greedy=False
)

optional_words = WordBoundary(space & NonCapturingGroup(Pattern(r"\w", count_min=0, greedy=False)) & space)

dates = date_prefix & optional_words & (eu_date | ind_date) & optional_words & times


def get_date(date_str):
    delimiters = (hyphen | colon | dot | backslash | space | ((space & comma & space) | comma))
    eu_y_m_d = NamedCapturingGroup("year", years) & delimiters \
               & NamedCapturingGroup("month", months) & delimiters \
               & NamedCapturingGroup("day", day_numbers) & delimiters & day_postfix

    in_d_m_y = NamedCapturingGroup("day", day_numbers) & delimiters & day_postfix & delimiters \
               & NamedCapturingGroup("month", months) & delimiters \
               & NamedCapturingGroup("year", years)

    grp_list = pp.Regex(dates.pattern, flags=IGNORECASE).scanString(date_str)

    if grp_list:  # if grp items is False skip it
        for text, start, end in grp_list:
            dt = {}
            if eu_date.name in text and any([text[eu_date.name]]):
                for txt, st, en in pp.Regex(eu_y_m_d.pattern, flags=IGNORECASE).scanString(text[eu_date.name]):
                    d = None
                    try:
                        month = int(txt["month"])
                        d = datetime(year=int(txt["year"]), month=month, day=int(txt["day"]))
                    except (Exception,):
                        try:
                            month = _monthsToNum[txt["month"][:3].lower()]
                            d = datetime(year=int(txt["year"]), month=month, day=int(txt["day"]))
                        except (Exception,):
                            pass
                    if d is not None:
                        dt.update(dict(datetime=d, **txt))

            if ind_date.name in text and any([text[ind_date.name]]):
                for txt, st, en in pp.Regex(in_d_m_y.pattern, flags=IGNORECASE).scanString(text[ind_date.name]):
                    d = None
                    try:
                        month = int(txt["month"])
                        d = datetime(year=int(txt["year"]), month=month, day=int(txt["day"]))
                    except (Exception,):
                        try:
                            month = _monthsToNum[txt["month"][:3].lower()]
                            d = datetime(year=int(txt["year"]), month=month, day=int(txt["day"]))
                        except (Exception,):
                            pass
                    if d is not None:
                        dt.update(dict(datetime=d, **txt))
            if any(dt):
                date_dict = dict(span=[start, end], **text, **dt)
                yield date_dict

#
# if __name__ == '__main__':
#     get_date("2020/feb/29")
#     get_date("date: 2021/01/31, 20/01/2001")
#     get_date("Date: 2021/12/31, 20/12/2001")
#     get_date("dated: 2021/12/31, 20/12/2001")
#     get_date("Dated: 2021/12/31, 20/12/2001")
#     get_date("dob: 2021/12/31, 20/12/2001")
#     get_date("Dob: 2021/12/31, 20/12/2001")
#     get_date("DoB: 2021/12/31, 20/12/2001")
#     get_date("DOB: 2021/12/31, 20/12/2001")
#     get_date("d.ob: 2021/12/31, 20/12/2001")
#     get_date("D.ob: 2021/12/31, 20/12/2001")
#     get_date("D.oB: 2021/12/31, 20/12/2001")
#     get_date("D.OB: 2021/12/31, 20/12/2001")
#     get_date("do.b: 2021/12/31, 20/12/2001")
#     get_date("Do.b: 2021/12/31, 20/12/2001")
#     get_date("Do.B: 2021/12/31, 20/12/2001")
#     get_date("DO.B: 2021/12/31, 20/12/2001")
#     get_date("d.o.b: 2021/12/31, 20/12/2001")
#     get_date("D.o.b: 2021/12/31, 20/12/2001")
#     get_date("D.o.B: 2021/12/31, 20/12/2001")
#     get_date("D.O.B: 2021/12/31, 20/12/2001")
#     get_date("D.O.BB: 2021/12/31, 20/12/2001")
#     get_date("D.O.BB: 2021/03/31, 20/12/2001")
#     get_date("D.O.BB: 2021/3/31, 20/12/2001")
#     get_date("D.O.BB: 2021/13/31, 20/12/2001")
#     get_date("2021/12/31, 20/12/2001")
#     get_date("2021/12/31, 20/June/2001")
#     get_date("2021/12/31, 20/june/2001")
#     get_date("2021/12/31, 20/jun/2001")
#     date = get_date("my date will be 2021/December/31, "
#                     " 2010 Apr 20 12 am "
#                     ""
#                     "on the date 20, APr, 2001 at 12:00 pm and many "
#                     "other dates when nations were born, 20-1-2031")
#     for dat in date:
#         print(dat)
#         print(dat["datetime"].isoformat())
#     print(*pp.Regex(dates.pattern, flags=IGNORECASE).scanString("Wedding date is fixed to be 20th, Apr, 2010"))
#     print(ind_date.alias)



# Copyright (c) 2022 Valio
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import typing
import unittest
from dataclasses import dataclass

from valio import (PatternValidator, ReassignValidator, RequiredValidator,
                   TypeValidator, __version__)
from valio.validator.validators import MultipleValidator


class TestVersion(unittest.TestCase):
    
    def test_version(self):
        self.assertEqual(__version__, '0.1.0b2')
        
        

class TestTypeValidator(unittest.TestCase):
    
    def setUp(self) -> None:
        self.validator = TypeValidator()
        self.debug_validator = TypeValidator(debug=True)
        
        def type_class_common_validator(typ, tp=None, tp_debug=None):
            @dataclass
            class TypeClass(object):
                tp: typing.Union[typ, TypeValidator, None] = self.validator
                tp_debug: typing.Union[typ, TypeValidator, None] = self.debug_validator
            return TypeClass(tp, tp_debug)

        def type_class(typ, tp=None, tp_debug=None):
            @dataclass
            class TypeClass(object):
                tp: typing.Union[typ, TypeValidator, None] = TypeValidator()
                tp_debug: typing.Union[typ, TypeValidator, None] = TypeValidator(debug=True)
            return TypeClass(tp, tp_debug)
        
        self.type_class = type_class
        self.type_class_common_validator = type_class_common_validator

    def test_type_class_validator(self):
        # typ = int, integer case
        self.assertIsNone(self.type_class(typ=int).tp)
        self.assertIsNone(self.type_class(typ=int).tp_debug)
        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        # with debug = False; the error is passed silently and value becomes   #
        # None                                                                 #
        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        self.assertIsNone(self.type_class(typ=int, tp="-1").tp)
        # ____________________________________________________________________ #
        
        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        # below assertEqual is passed because both isinstance(True, int) and   #
        # isinstance(False, int) are True                                      #
        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        self.assertEqual(self.type_class(typ=int, tp=True).tp, True)
        self.assertEqual(self.type_class(typ=int, tp=False).tp, False)
        self.assertEqual(self.type_class(typ=int, tp_debug=True).tp_debug, True)
        self.assertEqual(self.type_class(typ=int, tp_debug=False).tp_debug, False)
        # ____________________________________________________________________ # 
        
        with self.assertRaises(TypeError):
            self.type_class(typ=int, tp_debug="-1")
        
        self.assertEqual(self.type_class(typ=int, tp=~1).tp, ~1)
        self.assertEqual(self.type_class(typ=int, tp_debug=~1).tp_debug, ~1)
        # typ = bool, boolean case
        self.assertIsNone(self.type_class(typ=bool).tp)
        self.assertIsNone(self.type_class(typ=bool).tp_debug)
        self.assertIsNone(self.type_class(typ=bool, tp="-1").tp)
        with self.assertRaises(TypeError):
            self.type_class(typ=bool, tp_debug="-1")
        self.assertEqual(self.type_class(typ=bool, tp_debug=True).tp_debug, True)
        self.assertEqual(self.type_class(typ=bool, tp_debug=bool(~1)).tp_debug, bool(~1))
        self.assertEqual(self.type_class(typ=bool, tp_debug=False).tp_debug, False)
        # typ = Union[str, int]
        self.assertEqual(self.type_class(typ=typing.Union[str, int], tp="a").tp, "a")
        
        #      ____ COMMON VALIDATORS ARE REASSIGNED HERE BELOW ____
        # The expected behaviour is that common validators if re-assigned 
        # to a variable accross classes, works only if the variable in the 
        # reassigned class has same annotations or subclass of prior assigned 
        # annotation.
        # Below validation works with reassignment because issubclasss(bool, int)
        #
        # typ = int, integer case
        self.assertIsNone(self.type_class_common_validator(typ=int).tp)
        self.assertIsNone(self.type_class_common_validator(typ=int).tp_debug)
        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        # with debug = False; the error is passed silently and value becomes   #
        # None                                                                 #
        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        self.assertIsNone(self.type_class_common_validator(typ=int, tp="-1").tp)
        # ____________________________________________________________________ #
        
        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        # below assertEqual is passed because both isinstance(True, int) and   #
        # isinstance(False, int) are True                                      #
        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        self.assertEqual(self.type_class_common_validator(typ=int, tp=True).tp, True)
        self.assertEqual(self.type_class_common_validator(typ=int, tp=False).tp, False)
        self.assertEqual(self.type_class_common_validator(typ=int, tp_debug=True).tp_debug, True)
        self.assertEqual(self.type_class_common_validator(typ=int, tp_debug=False).tp_debug, False)
        # ____________________________________________________________________ # 
        
        with self.assertRaises(TypeError):
            self.type_class_common_validator(typ=int, tp_debug="-1")
        
        self.assertEqual(self.type_class_common_validator(typ=int, tp=~1).tp, ~1)
        self.assertEqual(self.type_class_common_validator(typ=int, tp_debug=~1).tp_debug, ~1)
        # typ = bool, boolean case
        self.assertIsNone(self.type_class_common_validator(typ=bool).tp)
        self.assertIsNone(self.type_class_common_validator(typ=bool).tp_debug)
        self.assertIsNone(self.type_class_common_validator(typ=bool, tp="-1").tp)
        with self.assertRaises(TypeError):
            self.type_class_common_validator(typ=bool, tp_debug="-1")
        self.assertEqual(self.type_class_common_validator(typ=bool, tp_debug=True).tp_debug, True)
        self.assertEqual(self.type_class_common_validator(typ=bool, tp_debug=bool(~1)).tp_debug, bool(~1))
        self.assertEqual(self.type_class_common_validator(typ=bool, tp_debug=False).tp_debug, False)
        
        #     ____________________________________________________________
        
        # CASE: typ = Union[str, int],
        # below case here fails because prior to reassignment of self.validator
        # it was annotated as 'bool' due to above test case and now when we reassign
        # the self.validator it checks if new annotation is same as 'bool'[in this case]
        # or is subclass of 'bool' [in this case], as the new tupe is Union[str, int] it fails
        with self.assertRaises(RuntimeError):
            self.assertEqual(self.type_class_common_validator(typ=typing.Union[str, int], tp="a").tp, "a")


class TestRequiredValidator(unittest.TestCase):
    
    def setUp(self) -> None:
        required_validator = typing.Union[str, RequiredValidator, None]
        self.required_validator: RequiredValidator = RequiredValidator
        self.required_empty: RequiredValidator = RequiredValidator()
        self.required_none: RequiredValidator = RequiredValidator(required=None)
        self.required_true: RequiredValidator = RequiredValidator(required=True, default="q")
        self.required_false: RequiredValidator = RequiredValidator(required=False)
        self.required_empty_debug_true: RequiredValidator = RequiredValidator(debug=True)
        self.required_none_debug_true: RequiredValidator = RequiredValidator(required=None, debug=True)
        self.required_true_debug_true: RequiredValidator = RequiredValidator(required=True, debug=True)
        self.required_false_debug_true: RequiredValidator = RequiredValidator(required=False, debug=True)
        
        @dataclass
        class TestRequiredClass(object):
            required_empty: required_validator = self.required_empty
            required_none: required_validator = self.required_none
            required_true: required_validator = self.required_true
            required_false: required_validator = self.required_false
            required_empty_debug_true: required_validator = self.required_empty_debug_true
            required_none_debug_true: required_validator = self.required_none_debug_true
            required_true_debug_true: required_validator = self.required_true_debug_true
            required_false_debug_true: required_validator = self.required_false_debug_true
            
        self.class_required = TestRequiredClass
        
    def test_required_is_valid(self):
        with self.assertRaises(KeyError):
            self.required_validator().required 
        with self.assertRaises(KeyError):
            self.required_validator(required=None).required 
        self.required_validator()
        self.required_validator(required=None)
        self.assertEqual(self.required_validator(required=True).required, True)
        self.assertEqual(self.required_validator(required=False).required, False)

    def test_required_is_invalid(self):
        # reasign value of invalid type
        with self.assertRaises(TypeError):
            self.required_validator(required="True")
        
        with self.assertRaises(TypeError):
            self.required_validator(required="False")
            
        with self.assertRaises(TypeError):
            self.required_validator(required=list)
            
        with self.assertRaises(TypeError):
            self.required_validator(required=TypeValidator(debug=True))
    
    def test_required_class(self):
        with self.assertRaises(ValueError):
            self.class_required(required_true_debug_true=None)
        self.class_required(required_true_debug_true="OK")
        
        @dataclass
        class RequiredEmptyDebugNone(object):
            value: RequiredValidator = RequiredValidator(debug=None)
        
        @dataclass
        class RequiredNoneDebugNone(object):
            value: RequiredValidator = RequiredValidator(required=None, debug=None)
            
        @dataclass
        class RequiredTrueDebugNone(object):
            value: RequiredValidator = RequiredValidator(required=True, debug=None)
        
        @dataclass
        class RequiredFalseDebugNone(object):
            value: RequiredValidator = RequiredValidator(required=False, debug=None)
        
        # value empty
        self.assertIsNone(RequiredEmptyDebugNone().value)
        self.assertIsNone(RequiredNoneDebugNone().value)
        self.assertIsNone(RequiredTrueDebugNone().value)
        self.assertIsNone(RequiredFalseDebugNone().value)
        
        # value is None
        self.assertIsNone(RequiredEmptyDebugNone(value=None).value)
        self.assertIsNone(RequiredNoneDebugNone(value=None).value)
        self.assertIsNone(RequiredTrueDebugNone(value=None).value)
        self.assertIsNone(RequiredFalseDebugNone(value=None).value)
        
        # value is False
        self.assertEqual(RequiredEmptyDebugNone(value=False).value, False)
        self.assertEqual(RequiredNoneDebugNone(value=False).value, False)
        self.assertEqual(RequiredTrueDebugNone(value=False).value, False)
        self.assertEqual(RequiredFalseDebugNone(value=False).value, False)
        
        # value is True
        self.assertEqual(RequiredEmptyDebugNone(value=True).value, True)
        self.assertEqual(RequiredNoneDebugNone(value=True).value, True)
        self.assertEqual(RequiredTrueDebugNone(value=True).value, True)
        self.assertEqual(RequiredFalseDebugNone(value=True).value, True)


        @dataclass
        class RequiredEmptyDebugFalse(object):
            value: RequiredValidator = RequiredValidator(debug=False)
        
        @dataclass
        class RequiredNoneDebugFalse(object):
            value: RequiredValidator = RequiredValidator(required=None, debug=False)
            
        @dataclass
        class RequiredTrueDebugFalse(object):
            value: RequiredValidator = RequiredValidator(required=True, debug=False)
        
        @dataclass
        class RequiredFalseDebugFalse(object):
            value: RequiredValidator = RequiredValidator(required=False, debug=False)
            
        
        # value is None
        self.assertIsNone(RequiredEmptyDebugFalse(value=None).value)
        self.assertIsNone(RequiredNoneDebugFalse(value=None).value)
        self.assertIsNone(RequiredTrueDebugFalse(value=None).value)
        self.assertIsNone(RequiredFalseDebugFalse(value=None).value)
        
        # value is False
        self.assertEqual(RequiredEmptyDebugFalse(value=False).value, False)
        self.assertEqual(RequiredNoneDebugFalse(value=False).value, False)
        self.assertEqual(RequiredTrueDebugFalse(value=False).value, False)
        self.assertEqual(RequiredFalseDebugFalse(value=False).value, False)
        
        # value is True
        self.assertEqual(RequiredEmptyDebugFalse(value=True).value, True)
        self.assertEqual(RequiredNoneDebugFalse(value=True).value, True)
        self.assertEqual(RequiredTrueDebugFalse(value=True).value, True)
        self.assertEqual(RequiredFalseDebugFalse(value=True).value, True)
        
        @dataclass
        class RequiredEmptyDebugTrue(object):
            value: RequiredValidator = RequiredValidator(debug=True)
        
        @dataclass
        class RequiredNoneDebugTrue(object):
            value: RequiredValidator = RequiredValidator(required=None, debug=True)
            
        @dataclass
        class RequiredTrueDebugTrue(object):
            value: RequiredValidator = RequiredValidator(required=True, debug=True)
        
        @dataclass
        class RequiredFalseDebugTrue(object):
            value: RequiredValidator = RequiredValidator(required=False, debug=True)
            
        
        # value is None
        self.assertIsNone(RequiredEmptyDebugTrue(value=None).value)
        self.assertIsNone(RequiredNoneDebugTrue(value=None).value)
        with self.assertRaises(ValueError):
            RequiredTrueDebugTrue(value=None).value
        self.assertIsNone(RequiredFalseDebugTrue(value=None).value)
        
        # value is False
        self.assertEqual(RequiredEmptyDebugTrue(value=False).value, False)
        self.assertEqual(RequiredNoneDebugTrue(value=False).value, False)
        self.assertEqual(RequiredTrueDebugTrue(value=False).value, False)
        self.assertEqual(RequiredFalseDebugTrue(value=False).value, False)
        
        # value is True
        self.assertEqual(RequiredEmptyDebugTrue(value=True).value, True)
        self.assertEqual(RequiredNoneDebugTrue(value=True).value, True)
        self.assertEqual(RequiredTrueDebugTrue(value=True).value, True)
        self.assertEqual(RequiredFalseDebugTrue(value=True).value, True)
        

class TestPatternValidator(unittest.TestCase):
    
    def setUp(self) -> None:
        self.word_1_or_more = r'\w+' 
        
    
    def test_pattern(self):
        @dataclass
        class PatternEmptyDebugNone(object):
            value: str | PatternValidator = PatternValidator()
            
        @dataclass
        class PatternEmptyDebugTrue(object):
            value: str | PatternValidator = PatternValidator(debug=True)
            
        @dataclass
        class PatternEmptyDebugFalse(object):
            value: str | PatternValidator = PatternValidator(debug=False)
        
        self.assertEqual(PatternEmptyDebugNone(value="1").value, "1")
        self.assertEqual(PatternEmptyDebugTrue(value="1").value, "1")  
        self.assertEqual(PatternEmptyDebugFalse(value="1").value, "1")
        @dataclass
        class PatternAndDebugNone(object):
            value: str | PatternValidator = PatternValidator(pattern=self.word_1_or_more)
            
        @dataclass
        class PatternAndDebugTrue(object):
            value: str | PatternValidator = PatternValidator(pattern=self.word_1_or_more, debug=True)
        
        @dataclass
        class PatternAndDebugFalse(object):
            value: str | PatternValidator = PatternValidator(pattern=self.word_1_or_more, debug=False)

        # value = None
        self.assertEqual(PatternAndDebugNone(value=None).value, None)
        self.assertEqual(PatternAndDebugFalse(value=None).value, None)
        self.assertEqual(PatternAndDebugTrue(value=None).value, None)
        
        # value = ""
        self.assertEqual(PatternAndDebugNone(value="").value, None)
        self.assertEqual(PatternAndDebugFalse(value="").value, None)
        with self.assertRaises(ValueError):
            PatternAndDebugTrue(value="").value
        
        # value = "a string"
        self.assertEqual(PatternAndDebugNone(value="a string").value, "a string")
        self.assertEqual(PatternAndDebugFalse(value="a string").value, "a string")
        self.assertEqual(PatternAndDebugTrue(value="a string").value, "a string")
        
class TestReassignValidator(unittest.TestCase):

    def setUp(self) -> None:
        self.reassign_validator: typing.Type[ReassignValidator] = ReassignValidator
        self.empty: ReassignValidator = ReassignValidator()
        self.reassign_true: ReassignValidator = ReassignValidator(reassign=True)
        self.reassign_false: ReassignValidator = ReassignValidator(reassign=False)
        self.reassign_none: ReassignValidator = ReassignValidator(reassign=None)
        self.reassign_empty_debug_true: ReassignValidator = ReassignValidator(debug=True)
        self.reassign_true_debug_true: ReassignValidator = ReassignValidator(reassign=True, debug=True)
        self.reassign_false_debug_true: ReassignValidator = ReassignValidator(reassign=False, debug=True)
        self.reassign_none_debug_true: ReassignValidator = ReassignValidator(reassign=None, debug=True)
        
        @dataclass
        class TestReassignClass(object):
            empty: typing.Union[str, ReassignValidator, None] =  self.empty
            true: typing.Union[str, ReassignValidator, None] =  self.reassign_true
            false: typing.Union[str, ReassignValidator, None] =  self.reassign_false
            none: typing.Union[str, ReassignValidator, None] =  self.reassign_none
            reassign_empty_with_debug: typing.Union[str, ReassignValidator, None] =  self.reassign_empty_debug_true
            reassign_true_with_debug: typing.Union[str, ReassignValidator, None] =  self.reassign_true_debug_true
            reassign_false_with_debug: typing.Union[str, ReassignValidator, None] =  self.reassign_false_debug_true
            reassign_none_with_debug: typing.Union[str, ReassignValidator, None] =  self.reassign_none_debug_true

        self.class_reassigned = TestReassignClass(
            empty="Reassign Empty Debug Empty", 
            true="Reassign True Debug Empty", 
            false="Reassign False Debug Empty", 
            none="Reassign None Debug Empty", 
            reassign_empty_with_debug="Reassign Empty Debug True", 
            reassign_true_with_debug="Reassign True Debug True", 
            reassign_false_with_debug="Reassign False Debug True", 
            reassign_none_with_debug="Reassign None Debug True"
            )
    
    def test_reassign_is_valid(self):
        with self.assertRaises(KeyError):
            self.reassign_validator().reassign
        with self.assertRaises(KeyError):
            self.reassign_validator(reassign=None).reassign
        self.reassign_validator()
        self.assertEqual(self.reassign_validator(reassign=True).reassign, True)    
        self.assertEqual(self.reassign_validator(reassign=False).reassign, False) 
        
    def test_reassign_is_invalid(self):
        # reasign value of invalid type
        with self.assertRaises(TypeError):
            self.reassign_validator(reassign="True")
        
        with self.assertRaises(TypeError):
            self.reassign_validator(reassign="False")
            
        with self.assertRaises(TypeError):
            self.reassign_validator(reassign=list)
            
        with self.assertRaises(TypeError):
            self.reassign_validator(reassign=TypeValidator(debug=True))
            
    def test_reassign_empty(self):
        self.assertIsNotNone(self.class_reassigned.empty)
        self.assertEqual(self.class_reassigned.empty, "Reassign Empty Debug Empty")
        self.assertGreater(self.empty.number_of_assignment, 0)
        self.assertEqual(self.empty.number_of_assignment, 1)
        self.assertGreater(self.empty.__dict__[id(self.class_reassigned)], 0)
        self.assertEqual(self.empty.__dict__[id(self.class_reassigned)], 1)
        self.class_reassigned.empty = "Reassigned"
        self.assertEqual(self.class_reassigned.empty, "Reassigned")
        self.assertGreater(self.empty.__dict__[id(self.class_reassigned)], 1)
        self.assertEqual(self.empty.__dict__[id(self.class_reassigned)], 2)
        
        # this should not fail because of TypeError as no Type Checks happen 
        # in ReassignValidator
        self.class_reassigned.empty = False  # notice empty type is 'str' but got set with 'bool'
        self.assertEqual(self.class_reassigned.empty, False)
        self.assertGreater(self.empty.number_of_assignment, 0)
        self.assertEqual(self.empty.number_of_assignment, 3)
        
    
    def test_reassign_none(self):
        self.assertIsNotNone(self.class_reassigned.none)
        self.assertEqual(self.class_reassigned.none, "Reassign None Debug Empty")
        self.assertGreater(self.reassign_none.number_of_assignment, 0)
        self.assertEqual(self.reassign_none.number_of_assignment, 1)
        self.class_reassigned.none = "Reassigned"
        self.assertEqual(self.class_reassigned.none, "Reassigned")
        self.assertGreater(self.reassign_none.number_of_assignment, 0)
        self.assertEqual(self.reassign_none.number_of_assignment, 2)
        
    def test_reassign_true(self):
        self.assertIsNotNone(self.class_reassigned.true)
        self.assertEqual(self.class_reassigned.true, "Reassign True Debug Empty")
        self.assertGreater(self.reassign_true.number_of_assignment, 0)
        self.assertEqual(self.reassign_true.number_of_assignment, 1)
        self.class_reassigned.true = "Reassigned"
        self.assertEqual(self.class_reassigned.true, "Reassigned")
        self.assertGreater(self.reassign_true.number_of_assignment, 0)
        self.assertEqual(self.reassign_true.number_of_assignment, 2)
        
    def test_reassign_false(self):
        self.assertIsNotNone(self.class_reassigned.false)
        self.assertEqual(self.class_reassigned.false, "Reassign False Debug Empty")
        self.assertGreater(self.reassign_false.number_of_assignment, 0)
        self.assertEqual(self.reassign_false.number_of_assignment, 1)
        self.class_reassigned.false = "Reassigned" # by default debug=None so no error raised but assignment fails by design
        self.assertNotEqual(self.class_reassigned.false, "Reassigned")
        self.assertEqual(self.class_reassigned.false, "Reassign False Debug Empty")
        self.assertGreater(self.reassign_false.number_of_assignment, 0)
        self.assertEqual(self.reassign_false.number_of_assignment, 1)
        
    def test_reassign_empty_with_debug(self):
        self.assertIsNotNone(self.class_reassigned.reassign_empty_with_debug)
        self.assertEqual(self.class_reassigned.reassign_empty_with_debug, "Reassign Empty Debug True")
        self.assertGreater(self.reassign_empty_debug_true.number_of_assignment, 0)
        self.assertEqual(self.reassign_empty_debug_true.number_of_assignment, 1)
        self.class_reassigned.reassign_empty_with_debug = "Reassigned"
        self.assertEqual(self.class_reassigned.reassign_empty_with_debug, "Reassigned")
        self.assertGreater(self.reassign_empty_debug_true.number_of_assignment, 0)
        self.assertEqual(self.reassign_empty_debug_true.number_of_assignment, 2)
        
        # this should not fail because of TypeError as no Type Checks happen 
        # in ReassignValidator
        self.class_reassigned.reassign_empty_with_debug = False  # notice debug_empty type is 'str' but got set with 'bool'
        self.assertEqual(self.class_reassigned.reassign_empty_with_debug, False)
        self.assertGreater(self.reassign_empty_debug_true.number_of_assignment, 0)
        self.assertEqual(self.reassign_empty_debug_true.number_of_assignment, 3)
        
    def test_reassign_none_with_debug(self):
        self.assertIsNotNone(self.class_reassigned.reassign_none_with_debug)
        self.assertEqual(self.class_reassigned.reassign_none_with_debug, "Reassign None Debug True")
        self.assertGreater(self.reassign_none_debug_true.number_of_assignment, 0)
        self.assertEqual(self.reassign_none_debug_true.number_of_assignment, 1)
        self.class_reassigned.reassign_none_with_debug = "Reassigned"
        self.assertEqual(self.class_reassigned.reassign_none_with_debug, "Reassigned")
        self.assertGreater(self.reassign_none_debug_true.number_of_assignment, 0)
        self.assertEqual(self.reassign_none_debug_true.number_of_assignment, 2)
        
        # this should not fail because of TypeError as no Type Checks happen 
        # in ReassignValidator
        self.class_reassigned.reassign_none_with_debug = False  # notice debug_none type is 'str' but got set with 'bool'
        self.assertEqual(self.class_reassigned.reassign_none_with_debug, False)
        self.assertGreater(self.reassign_none_debug_true.number_of_assignment, 0)
        self.assertEqual(self.reassign_none_debug_true.number_of_assignment, 3)
        
    def test_reassign_true_with_debug(self):
        self.assertIsNotNone(self.class_reassigned.reassign_true_with_debug)
        self.assertEqual(self.class_reassigned.reassign_true_with_debug, "Reassign True Debug True")
        self.assertGreater(self.reassign_true_debug_true.number_of_assignment, 0)
        self.assertEqual(self.reassign_true_debug_true.number_of_assignment, 1)
        self.class_reassigned.reassign_true_with_debug = "Reassigned"
        self.assertEqual(self.class_reassigned.reassign_true_with_debug, "Reassigned")
        self.assertGreater(self.reassign_true_debug_true.number_of_assignment, 0)
        self.assertEqual(self.reassign_true_debug_true.number_of_assignment, 2)
        
        # this should not fail because of TypeError as no Type Checks happen 
        # in ReassignValidator
        self.class_reassigned.reassign_true_with_debug = False  # <- notice no type validation, expects 'str' but set with 'bool'
        self.assertEqual(self.class_reassigned.reassign_true_with_debug, False)
        self.assertGreater(self.reassign_true_debug_true.number_of_assignment, 0)
        self.assertEqual(self.reassign_true_debug_true.number_of_assignment, 3)

    def test_reassign_false_with_debug(self):
        self.assertIsNotNone(self.class_reassigned.reassign_false_with_debug)
        self.assertEqual(self.class_reassigned.reassign_false_with_debug, "Reassign False Debug True")
        self.assertGreater(self.reassign_false_debug_true.number_of_assignment, 0)
        self.assertEqual(self.reassign_false_debug_true.number_of_assignment, 1)
        with self.assertRaises(AttributeError): 
            self.class_reassigned.reassign_false_with_debug = "Reassigned"
        self.assertNotEqual(self.class_reassigned.reassign_false_with_debug, "Reassigned")
        self.assertEqual(self.class_reassigned.reassign_false_with_debug, "Reassign False Debug True")
        self.assertGreater(self.reassign_false_debug_true.number_of_assignment, 0)
        self.assertEqual(self.reassign_false_debug_true.number_of_assignment, 1)
        
        # this should fail not because of TypeError as no Type Checks happen 
        # in ReassignValidator for values assigned but due to reassign_validation
        with self.assertRaises(AttributeError):
            # notice no type validation, expects 'str' but set with 'bool' error due to reassign_validation raising
            # AttributeError
            self.class_reassigned.reassign_false_with_debug = False  
        self.assertNotEqual(self.class_reassigned.reassign_false_with_debug, False)
        self.assertEqual(self.class_reassigned.reassign_false_with_debug, "Reassign False Debug True")
        self.assertGreater(self.reassign_false_debug_true.number_of_assignment, 0)
        self.assertEqual(self.reassign_false_debug_true.number_of_assignment, 1)
        
class TestMultipleValidator(unittest.TestCase):
    
    def setUp(self) -> None:
        self.multiple_of = 2
    
    def test_multiple_of_validator(self):
        @dataclass
        class MultipleEmptyDebugEmpty(object):
            value: int| MultipleValidator | None = MultipleValidator()
            
        @dataclass
        class MultipleEmptyDebugTrue(object):
            value: int| MultipleValidator | None = MultipleValidator(debug=True)
            
        @dataclass
        class MultipleEmptyDebugFalse(object):
            value: int| MultipleValidator | None = MultipleValidator(debug=False)
            
        self.assertEqual(MultipleEmptyDebugEmpty(value=None).value, None)
        self.assertEqual(MultipleEmptyDebugEmpty(value="Abc").value, 'Abc') # no typechecks
        self.assertEqual(MultipleEmptyDebugEmpty(value=1).value, 1)            

if __name__ == '__main__':
    unittest.main()
    




# Copyright (c) 2022 Valio
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT



import io
import logging
import logging.handlers
import os
import re
import sys
from typing import IO, List, Optional, Tuple

from typing_extensions import Final, Literal

try:
    import curses

    import _curses  # noqa

    CURSES_ENABLED = True
except ImportError:
    CURSES_ENABLED = False

# so we can't get the color code from curses.
PLAIN_ANSI_DIM = "\x1b[2m"  # type: Final

DEFAULT_SOURCE_OFFSET = 4  # type: Final
DEFAULT_COLUMNS = 80  # type: Final

# At least this number of columns will be shown on each side of
# error location when printing source code snippet.
MINIMUM_WIDTH = 20

# VT100 color code processing was added in Windows 10, but only the second major update,
# Threshold 2. Fortunately, everyone (even on LTSB, Long Term Support Branch) should
# have a version of Windows 10 newer than this. Note that Windows 8 and below are not
# supported, but are either going out of support, or make up only a few % of the market.
MINIMUM_WINDOWS_MAJOR_VT100 = 10
MINIMUM_WINDOWS_BUILD_VT100 = 10586


class ColorCodes:
    grey = "\x1b[38;21m"
    green = "\x1b[1;32m"
    yellow = "\x1b[33;21m"
    red = "\x1b[31;21m"
    bold_red = "\x1b[31;1m"
    blue = "\x1b[1;34m"
    light_blue = "\x1b[1;36m"
    purple = "\x1b[1;35m"
    reset = "\x1b[0m"


class ColorizedArgsFormatter(logging.Formatter):
    """
    refer : https://stackoverflow.com/a/61996622
    Usage:
        root_logger = logging.getLogger()
        console_handler = logging.StreamHandler(stream=sys.stdout)
        console_format = "%(asctime)s - %(levelname)-8s - %(name)-25s - %(message)s"
        colored_formatter = ColorizedArgsFormatter(console_format)
        console_handler.setFormatter(colored_formatter)
        root_logger.addHandler(console_handler)

        logger = logging.getLogger(__name__)
        logger.info("Hello World")
        logger.info("Request from {} handled in {:.3f} ms", socket.gethostname(), 11)
        logger.info("Request from {} handled in {:.3f} ms", "127.0.0.1", 33.1)
        logger.info("My favorite drinks are {}, {}, {}, {}", "milk", "wine", "tea", "beer")
        logger.debug("this is a {} message", logging.getLevelName(logging.DEBUG))
        logger.info("this is a {} message", logging.getLevelName(logging.INFO))
        logger.warning("this is a {} message", logging.getLevelName(logging.WARNING))
        logger.error("this is a {} message", logging.getLevelName(logging.ERROR))
        logger.critical("this is a {} message", logging.getLevelName(logging.CRITICAL))
        logger.info("Does old-style formatting also work? %s it is, but no colors (yet)", True)

    """

    arg_colors = [ColorCodes.purple, ColorCodes.light_blue]
    level_fields = ["levelname", "levelno"]
    level_to_color = {
        logging.DEBUG: ColorCodes.grey,
        logging.INFO: ColorCodes.green,
        logging.WARNING: ColorCodes.yellow,
        logging.ERROR: ColorCodes.red,
        logging.CRITICAL: ColorCodes.bold_red,
    }

    def __init__(self, fmt: str):
        super().__init__()
        self.level_to_formatter = {}

        def add_color_format(level: int):
            color = ColorizedArgsFormatter.level_to_color[level]
            _format = fmt
            for fld in ColorizedArgsFormatter.level_fields:
                search = "(%\(" + fld + "\).*?s)"
                _format = re.sub(search, f"{color}\\1{ColorCodes.reset}", _format)
            formatter = logging.Formatter(_format)
            self.level_to_formatter[level] = formatter

        add_color_format(logging.DEBUG)
        add_color_format(logging.INFO)
        add_color_format(logging.WARNING)
        add_color_format(logging.ERROR)
        add_color_format(logging.CRITICAL)

    @staticmethod
    def rewrite_record(record: logging.LogRecord):
        if not BraceFormatStyleFormatter.is_brace_format_style(record):
            return

        msg = record.msg
        msg = msg.replace("{", "_{{")
        msg = msg.replace("}", "_}}")
        placeholder_count = 0
        # add ANSI escape code for next alternating color before each formatting parameter
        # and reset color after it.
        while True:
            if "_{{" not in msg:
                break
            color_index = placeholder_count % len(ColorizedArgsFormatter.arg_colors)
            color = ColorizedArgsFormatter.arg_colors[color_index]
            msg = msg.replace("_{{", color + "{", 1)
            msg = msg.replace("_}}", "}" + ColorCodes.reset, 1)
            placeholder_count += 1

        record.msg = msg.format(*record.args)
        record.args = []

    def format(self, record):
        orig_msg = record.msg
        orig_args = record.args
        formatter = self.level_to_formatter.get(record.levelno)
        self.rewrite_record(record)
        formatted = formatter.format(record)

        # restore log record to original state for other handlers
        record.msg = orig_msg
        record.args = orig_args
        return formatted


class BraceFormatStyleFormatter(logging.Formatter):
    def __init__(self, fmt: str):
        super().__init__()
        self.formatter = logging.Formatter(fmt)

    @staticmethod
    def is_brace_format_style(record: logging.LogRecord):
        if len(record.args) == 0:
            return False

        msg = record.msg
        if "%" in msg:
            return False

        count_of_start_param = msg.count("{")
        count_of_end_param = msg.count("}")

        if count_of_start_param != count_of_end_param:
            return False

        if count_of_start_param != len(record.args):
            return False

        return True

    @staticmethod
    def rewrite_record(record: logging.LogRecord):
        if not BraceFormatStyleFormatter.is_brace_format_style(record):
            return

        record.msg = record.msg.format(*record.args)
        record.args = []

    def format(self, record):
        orig_msg = record.msg
        orig_args = record.args
        self.rewrite_record(record)
        formatted = self.formatter.format(record)

        # restore log record to original state for other handlers
        record.msg = orig_msg
        record.args = orig_args
        return formatted


def split_words(msg: str) -> List[str]:
    """Split line of text into words (but not within quoted groups)."""
    next_word = ""
    res = []  # type: List[str]
    allow_break = True
    for c in msg:
        if c == " " and allow_break:
            res.append(next_word)
            next_word = ""
            continue
        if c == '"':
            allow_break = not allow_break
        next_word += c
    res.append(next_word)
    return res


def get_terminal_width() -> int:
    """Get current terminal width if possible, otherwise return the default one."""
    try:
        cols, _ = os.get_terminal_size()
    except OSError:
        return DEFAULT_COLUMNS
    else:
        if cols == 0:
            return DEFAULT_COLUMNS
        return cols


def soft_wrap(msg: str, max_len: int, first_offset: int, num_indent: int = 0) -> str:
    """Wrap a long error message into few lines.

    Breaks will only happen between words, and never inside a quoted group
    (to avoid breaking types such as "Union[int, str]"). The 'first_offset' is
    the width before the start of first line.

    Pad every next line with 'num_indent' spaces. Every line will be at most 'max_len'
    characters, except if it is a single word or quoted group.

    For example:
               first_offset
        ------------------------
        path/to/file: error: 58: Some very long error message
            that needs to be split in separate lines.
            "Long[Type, Names]" are never split.
        ^^^^--------------------------------------------------
        num_indent           max_len
    """
    words = split_words(msg)
    next_line = words.pop(0)
    lines = []  # type: List[str]
    while words:
        next_word = words.pop(0)
        max_line_len = max_len - num_indent if lines else max_len - first_offset
        # Add 1 to account for space between words.
        if len(next_line) + len(next_word) + 1 <= max_line_len:
            next_line += " " + next_word
        else:
            lines.append(next_line)
            next_line = next_word
    lines.append(next_line)
    padding = "\n" + " " * num_indent
    return padding.join(lines)


def trim_source_line(
        line: str, max_len: int, col: int, min_width: int
) -> Tuple[str, int]:
    """Trim a line of source code to fit into max_len.

    Show 'min_width' characters on each side of 'col' (an error location). If either
    start or end is trimmed, this is indicated by adding '...' there.
    A typical result looks like this:
        ...some_variable = function_to_call(one_arg, other_arg) or...

    Return the trimmed string and the column offset to to adjust error location.
    """
    if max_len < 2 * min_width + 1:
        # In case the window is too tiny it is better to still show something.
        max_len = 2 * min_width + 1

    # Trivial case: line already fits in.
    if len(line) <= max_len:
        return line, 0

    # If column is not too large so that there is still min_width after it,
    # the line doesn't need to be trimmed at the start.
    if col + min_width < max_len:
        return line[:max_len] + "...", 0

    # Otherwise, if the column is not too close to the end, trim both sides.
    if col < len(line) - min_width - 1:
        offset = col - max_len + min_width + 1
        return "..." + line[offset: col + min_width + 1] + "...", offset - 3

    # Finally, if the column is near the end, just trim the start.
    return "..." + line[-max_len:], len(line) - max_len - 3


class FancyFormatter:
    """Apply color and bold font to terminal output.

    This currently only works on Linux and Mac.
    """

    def __init__(self, f_out: IO[str], f_err: IO[str], show_error_codes: bool) -> None:
        self.show_error_codes = show_error_codes
        # Check if we are in a human-facing terminal on a supported platform.
        if sys.platform not in ("linux", "darwin", "win32"):
            self.dummy_term = True
            return
        force_color = int(os.getenv("MYPY_FORCE_COLOR", "0"))
        if not force_color and (not f_out.isatty() or not f_err.isatty()):
            self.dummy_term = True
            return
        if sys.platform == "win32":
            self.dummy_term = not self.initialize_win_colors()
        else:
            self.dummy_term = not self.initialize_unix_colors()
        if not self.dummy_term:
            self.colors = {
                "red": self.RED,
                "green": self.GREEN,
                "blue": self.BLUE,
                "yellow": self.YELLOW,
                "none": "",
            }

    def initialize_win_colors(self) -> bool:
        """Return True if initialization was successful and we can use colors, False otherwise"""
        # Windows ANSI escape sequences are only supported on Threshold 2 and above.
        # we check with an assert at runtime and an if check for mypy, as asserts do not
        # yet narrow platform
        assert sys.platform == "win32"
        if sys.platform == "win32":
            winver = sys.getwindowsversion()
            if (
                    winver.major < MINIMUM_WINDOWS_MAJOR_VT100
                    or winver.build < MINIMUM_WINDOWS_BUILD_VT100
            ):
                return False
            import ctypes

            kernel32 = ctypes.windll.kernel32
            ENABLE_PROCESSED_OUTPUT = 0x1
            ENABLE_WRAP_AT_EOL_OUTPUT = 0x2
            ENABLE_VIRTUAL_TERMINAL_PROCESSING = 0x4
            STD_OUTPUT_HANDLE = -11
            kernel32.SetConsoleMode(
                kernel32.GetStdHandle(STD_OUTPUT_HANDLE),
                ENABLE_PROCESSED_OUTPUT
                | ENABLE_WRAP_AT_EOL_OUTPUT
                | ENABLE_VIRTUAL_TERMINAL_PROCESSING,
            )
            self.BOLD = "\033[1m"
            self.UNDER = "\033[4m"
            self.BLUE = "\033[94m"
            self.GREEN = "\033[92m"
            self.RED = "\033[91m"
            self.YELLOW = "\033[93m"
            self.NORMAL = "\033[0m"
            self.DIM = "\033[2m"
            return True
        return False

    def initialize_unix_colors(self) -> bool:
        """Return True if initialization was successful and we can use colors, False otherwise"""
        if not CURSES_ENABLED:
            return False
        try:
            # setupterm wants a fd to potentially write an "initialization sequence".
            # We override sys.stdout for the daemon API so if stdout doesn't have an fd,
            # just give it /dev/null.
            try:
                fd = sys.stdout.fileno()
            except io.UnsupportedOperation:
                with open("/dev/null", "rb") as f:
                    curses.setupterm(fd=f.fileno())
            else:
                curses.setupterm(fd=fd)
        except curses.error:
            # Most likely terminfo not found.
            return False
        bold = curses.tigetstr("bold")
        under = curses.tigetstr("smul")
        set_color = curses.tigetstr("setaf")
        if not (bold and under and set_color):
            return False

        self.NORMAL = curses.tigetstr("sgr0").decode()
        self.BOLD = bold.decode()
        self.UNDER = under.decode()
        dim = curses.tigetstr("dim")
        # TODO: more reliable way to get gray color good for both dark and light schemes.
        self.DIM = dim.decode() if dim else PLAIN_ANSI_DIM

        self.BLUE = curses.tparm(set_color, curses.COLOR_BLUE).decode()
        self.GREEN = curses.tparm(set_color, curses.COLOR_GREEN).decode()
        self.RED = curses.tparm(set_color, curses.COLOR_RED).decode()
        self.YELLOW = curses.tparm(set_color, curses.COLOR_YELLOW).decode()
        return True

    def style(
            self,
            text: str,
            color: Literal["red", "green", "blue", "yellow", "none"],
            bold: bool = False,
            underline: bool = False,
            dim: bool = False,
    ) -> str:
        """Apply simple color and style (underlined or bold)."""
        if self.dummy_term:
            return text
        if bold:
            start = self.BOLD
        else:
            start = ""
        if underline:
            start += self.UNDER
        if dim:
            start += self.DIM
        return start + self.colors[color] + text + self.NORMAL

    def fit_in_terminal(
            self, messages: List[str], fixed_terminal_width: Optional[int] = None
    ) -> List[str]:
        """Improve readability by wrapping error messages and trimming source code."""
        width = (
                fixed_terminal_width
                or int(os.getenv("MYPY_FORCE_TERMINAL_WIDTH", "0"))
                or get_terminal_width()
        )
        new_messages = messages.copy()
        for i, error in enumerate(messages):
            if ": error:" in error:
                loc, msg = error.split("error:", maxsplit=1)
                msg = soft_wrap(msg, width, first_offset=len(loc) + len("error: "))
                new_messages[i] = loc + "error:" + msg
            if error.startswith(" " * DEFAULT_SOURCE_OFFSET) and "^" not in error:
                # TODO: detecting source code highlights through an indent can be surprising.
                # Restore original error message and error location.
                error = error[DEFAULT_SOURCE_OFFSET:]
                column = messages[i + 1].index("^") - DEFAULT_SOURCE_OFFSET

                # Let source have some space also on the right side, plus 6
                # to accommodate ... on each side.
                max_len = width - DEFAULT_SOURCE_OFFSET - 6
                source_line, offset = trim_source_line(
                    error, max_len, column, MINIMUM_WIDTH
                )

                new_messages[i] = " " * DEFAULT_SOURCE_OFFSET + source_line
                # Also adjust the error marker position.
                new_messages[i + 1] = (
                        " " * (DEFAULT_SOURCE_OFFSET + column - offset) + "^"
                )
        return new_messages

    def colorize(self, error: str) -> str:
        """Colorize an output line by highlighting the status and error code.

        If fixed_terminal_width is given, use it instead of calling get_terminal_width()
        (used by the daemon).
        """
        if ": error:" in error:
            loc, msg = error.split("error:", maxsplit=1)
            if not self.show_error_codes:
                return (
                        loc
                        + self.style("error:", "red", bold=True)
                        + self.highlight_quote_groups(msg)
                )
            codepos = msg.rfind("[")
            if codepos != -1:
                code = msg[codepos:]
                msg = msg[:codepos]
            else:
                code = ""  # no error code specified
            return (
                    loc
                    + self.style("error:", "red", bold=True)
                    + self.highlight_quote_groups(msg)
                    + self.style(code, "yellow")
            )
        elif ": note:" in error:
            loc, msg = error.split("note:", maxsplit=1)
            return loc + self.style("note:", "blue") + self.underline_link(msg)
        elif error.startswith(" " * DEFAULT_SOURCE_OFFSET):
            # TODO: detecting source code highlights through an indent can be surprising.
            if "^" not in error:
                return self.style(error, "none", dim=True)
            return self.style(error, "red")
        else:
            return error

    def highlight_quote_groups(self, msg: str) -> str:
        """Make groups quoted with double quotes bold (including quotes).

        This is used to highlight types, attribute names etc.
        """
        if msg.count('"') % 2:
            # Broken error message, don't do any formatting.
            return msg
        parts = msg.split('"')
        out = ""
        for i, part in enumerate(parts):
            if i % 2 == 0:
                out += self.style(part, "none")
            else:
                out += self.style('"' + part + '"', "none", bold=True)
        return out

    def underline_link(self, note: str) -> str:
        """Underline a link in a note message (if any).

        This assumes there is at most one link in the message.
        """
        match = re.search(r"https?://\S*", note)
        if not match:
            return note
        start = match.start()
        end = match.end()
        return (
                note[:start]
                + self.style(note[start:end], "none", underline=True)
                + note[end:]
        )

    def format_success(self, n_sources: int, use_color: bool = True) -> str:
        """Format short summary in case of success.

        n_sources is total number of files passed directly on command line,
        i.e. excluding stubs and followed imports.
        """
        msg = "Success: no issues found in {}" " source file{}".format(
            n_sources, "s" if n_sources != 1 else ""
        )
        if not use_color:
            return msg
        return self.style(msg, "green", bold=True)

    def format_error(
            self, n_errors: int, n_files: int, n_sources: int, use_color: bool = True
    ) -> str:
        """Format a short summary in case of errors."""
        msg = "Found {} error{} in {} file{}" " (checked {} source file{})".format(
            n_errors,
            "s" if n_errors != 1 else "",
            n_files,
            "s" if n_files != 1 else "",
            n_sources,
            "s" if n_sources != 1 else "",
        )
        if not use_color:
            return msg
        return self.style(msg, "red", bold=True)



# Copyright (c) 2022 Valio
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT



from functools import reduce
from pprint import pformat
from typing import Union

__all__ = [
    "PatternType",
    "Pattern",
    "All",
    "Any",
    "SetOf",
    "Escape",
    "CapturingGroup",
    "NonCapturingGroup",
    "NamedCapturingGroup",
    "IfEarlierNamedCapturedGroup",
    "Comment",
    "StartsWith",
    "EndsWith",
    "StartOfString",
    "EndOfString",
    "WordBoundary",
    "IfPrecededBy",
    "IfNotPrecededBy",
    "IfFollowedBy",
    "IfNotFollowedBy",
    "IfGroupNameMatched",
]


class PatternType(object):
    def __init__(
            self,
            pattern: Union[str, bytes, "PatternType", "Pattern", None, object],
            alias: str = None,
    ):
        self.pattern = (
            pattern.pattern
            if hasattr(pattern, "pattern")
            else pattern
            if isinstance(pattern, (str, bytes))
            else None
        )
        self.raw_pattern = (
            pattern.raw_pattern if hasattr(pattern, "raw_pattern") else self.pattern
        )
        self.quantifier = pattern.quantifier if hasattr(pattern, "quantifier") else fr""
        self.alias = alias or pattern.alias if hasattr(pattern, "alias") else self.alias \
            if hasattr(self, "alias") and any(self.alias) else fr""

    def __and__(self, other):
        if not isinstance(other, PatternType):
            raise TypeError(
                f"expected {type(self).__name__} type, "
                f"got {type(other).__name__} instead"
            )
        if all(
                [
                    isinstance(self.pattern, PatternType),
                    isinstance(other.pattern, PatternType),
                ]
        ):
            return AND(self.pattern, other.pattern)
        return AND(self, other)

    def __or__(self, other):
        if not isinstance(other, PatternType):
            raise TypeError(
                f"expected {type(self).__name__} type, "
                f"got {type(other).__name__} instead"
            )
        if all(
                [
                    isinstance(self.pattern, PatternType),
                    isinstance(other.pattern, PatternType),
                ]
        ):
            return OR(self.pattern, other.pattern)
        return OR(self, other)

    def __neg__(self):
        return NOT(self)

    def __iter__(self):
        for _ in [self.pattern]:
            yield _

    def __str__(self):
        return self.pattern

    def __repr__(self):
        return pformat(fr"<{__name__}.PatternType({repr(self.pattern)}) object at {hex(hash(id(self)))}>")

    def __set_name__(self, owner, name):
        if self.alias is None or not any([self.alias]):
            self.alias = name


class OR(PatternType):
    def __init__(
            self,
            first_pattern: PatternType,
            second_pattern: PatternType,
    ):
        _pattern = Pattern(f"{first_pattern.pattern}|{second_pattern.pattern}") if not isinstance(first_pattern,
                                                                                                  OR) else Pattern(
            f"{first_pattern.raw_pattern}|{second_pattern.pattern}")
        self.pattern = NonCapturingGroup(_pattern).pattern
        self.raw_pattern = _pattern.raw_pattern
        self.allow_set = not any([first_pattern.quantifier, second_pattern.quantifier])
        self.alias = " or ".join([first_pattern.alias, second_pattern.alias])
        super(OR, self).__init__(self)


class AND(PatternType):
    def __init__(
            self,
            first_pattern: PatternType,
            second_pattern: PatternType,
    ):
        self.pattern = f"{first_pattern.pattern}{second_pattern.pattern}"
        self.allow_set = not any([first_pattern.quantifier, second_pattern.quantifier])
        self.alias = "".join([first_pattern.alias, second_pattern.alias])
        super(AND, self).__init__(self)


class NOT(PatternType):
    def __init__(self, pattern: Union[PatternType, "SetOf"]):
        self.pattern = (
            f"!{pattern.pattern}"
            if not isinstance(pattern, SetOf)
            else (~pattern).pattern
        )
        super(NOT, self).__init__(self)


class Value(object):
    def __init__(self, min_value: int = None, max_value: int = None):
        _min_value = min_value or 0
        _max_value = max_value or _min_value
        if _max_value < _min_value:
            raise ValueError(
                f"expect max_value to be equal to or greater than {_min_value}, got {_max_value} instead"
            )

        self._min = min_value
        self._max = max_value
        self._value = None

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if self._min is not None:
            if value < self._min:
                raise ValueError(
                    f"expect value to be equal to or greater than {self._min}, got {value} instead"
                )
        if self._max is not None:
            if value > self._max:
                raise ValueError(
                    f"expect value to be equal to or less than {self._max}, got {value} instead"
                )
        self._value = value


class Quantifier(PatternType):
    def __init__(
            self,
            count_min: int = None,
            count_max: int = None,
            greedy: bool = True,
    ):
        if count_min is not None:
            if count_min < 0:
                raise ValueError(
                    f"expect count_min to be equal to or greater than 0, got {count_min} instead"
                )
        positive = Value(min_value=count_min or 0)
        _min = positive.value = count_min or 0
        _max = positive.value = count_max or _min
        del positive
        count = (
            fr""
            if all([count_min is None, count_max is None])
            else (
                f"{{{_min},{_max if count_max is not None else ''}}}"
                if count_min != count_max
                else f"{{{count_min}}}"
                if count_min is not None
                else f"{{{_min},{_max if count_max is not None else ''}}}"
            )
        )
        count_map = {f"{{0,}}": f"*", f"{{1,}}": f"+", f"{{0,1}}": f"?"}
        try:
            self.quantifier = (
                f"{count_map[count]}" if greedy else f"{count_map[count]}?"
            )
        except KeyError:
            self.quantifier = fr"{count}" if greedy else fr"{count}?"
        super(Quantifier, self).__init__(self)


class Pattern(PatternType):
    def __init__(
            self,
            /,  # NOQA
            pattern: Union[str, bytes],
            raw_pattern: Union[str, bytes, None] = None,
            quantifier: str = None,
            count: int = None,
            count_min: int = None,
            count_max: int = None,
            greedy: bool = True,
            alias: str = None,
    ):
        if any([all([count, count_min]), all([count, count_max])]):
            raise ValueError(
                f"count and count_min or count and count_max "
                f"values can not be used together"
            )
        quantifier = (
                quantifier
                or Quantifier(
            count_min=count_min or count if count is not None else count_min,
            count_max=count_max or count if count is not None else count_max, greedy=greedy
        ).quantifier
        )
        self.raw_pattern = raw_pattern if raw_pattern is not None else pattern
        self.quantifier = quantifier
        self.pattern = fr"{pattern}{quantifier}"
        super(Pattern, self).__init__(self)
        if alias is not None:
            self.alias = alias


class All(PatternType):
    def __init__(self, *patterns: PatternType):
        self.pattern = (
            reduce(lambda x, y: x & y, iter(patterns)) if any(patterns) else ""
        )
        super(All, self).__init__(self.pattern)


class Any(PatternType):
    def __init__(self, *patterns: PatternType):
        self.pattern = (
            reduce(lambda x, y: (x | y), iter(patterns)) if any(patterns) else f"."
        )
        super(Any, self).__init__(self.pattern)


class SetOf(PatternType):
    def __init__(
            self,
            pattern: Union[PatternType, Pattern, "SetOf"],
            count: int = None,
            count_min: int = None,
            count_max: int = None,
            greedy: bool = True,
            alias: str = None
    ):
        if any([all([count, count_min]), all([count, count_max])]):
            raise ValueError(
                f"count and count_min or count and count_max "
                f"values can not be used together"
            )
        if isinstance(pattern, (OR, AND)):
            if not pattern.allow_set:
                raise SyntaxError(
                    "expected SetOf(A) * SetOf(B) syntax, got SetOf(A*B) syntax instead"
                )
        set_quantifier = Quantifier(
            count_min=count_min or count if count is not None else count_min,
            count_max=count_max or count if count is not None else count_max, greedy=greedy
        ).quantifier
        self.raw_pattern = pattern.raw_pattern
        if any(set_quantifier) and any(pattern.quantifier):
            raise ValueError(
                fr"pattern quantifiers {pattern.quantifier} already exist, "
                fr"{SetOf.__name__} quantifiers can't be set"
            )
        self.quantifier = (
            set_quantifier if not any(pattern.quantifier) else pattern.quantifier
        )
        negate = pattern.set_negated if hasattr(pattern, "set_negated") else False
        self.pattern = (
            f"[{self.raw_pattern}]{self.quantifier}"
            if negate is None or not negate
            else f"[^{self.raw_pattern}]{self.quantifier}"
        )
        self.set_negated = negate
        super(SetOf, self).__init__(self)

    def __invert__(self):
        self.set_negated = ~self.set_negated
        return SetOf(self)


class Escape(PatternType):
    def __new__(cls, *args, **kwargs):
        if not any(kwargs) and len(args) == 1 and isinstance(args[0], SetOf):
            return ~args[0]
        __class__ = super(Escape, cls).__new__(*args, **kwargs)
        return __class__

    def __init__(self, pattern: PatternType):
        self.pattern = f"\\{pattern.raw_pattern}"
        self.raw_pattern = pattern.raw_pattern
        super(Escape, self).__init__(self)

    def __or__(self, other):
        if isinstance(other, Escape):
            return Escape(Pattern(f"{self.raw_pattern}|{other.raw_pattern}"))
        else:
            super(Escape, self).__or__(other)

    def __and__(self, other):
        if isinstance(other, Escape):
            return Escape(Pattern(f"{self.raw_pattern}{other.raw_pattern}"))
        else:
            super(Escape, self).__or__(other)


class CapturingGroup(PatternType):
    def __init__(
            self,
            pattern: PatternType,
            count_min: int = None,
            count: int = None,
            count_max: int = None,
            greedy: bool = True,
    ):
        self.pattern = Pattern(
            pattern=fr"({pattern.pattern})",
            count_min=count_min,
            count=count,
            count_max=count_max,
            greedy=greedy,
        )
        self.alias = pattern.alias
        super(CapturingGroup, self).__init__(self.pattern)


class NonCapturingGroup(PatternType):
    def __init__(
            self,
            pattern: PatternType,
            count_min: int = None,
            count: int = None,
            count_max: int = None,
            greedy: bool = True,
    ):
        self.pattern = Pattern(
            pattern=fr"(?:{pattern.pattern if not isinstance(pattern, OR) else pattern.raw_pattern})",
            count_min=count_min,
            count=count,
            count_max=count_max,
            greedy=greedy,
        )
        self.raw_pattern = pattern.raw_pattern
        self.alias = pattern.alias
        super(NonCapturingGroup, self).__init__(self.pattern)


class NamedCapturingGroup(PatternType):
    def __init__(
            self,
            name: str,
            pattern: PatternType,
            count_min: int = None,
            count: int = None,
            count_max: int = None,
            greedy: bool = True,
    ):
        if any([all([count, count_min]), all([count, count_max])]):
            raise ValueError(
                f"count and count_min or count and count_max "
                f"values can not be used together"
            )

        self.quantifier = Quantifier(
            count_min=count_min or count if count is not None else count_min,
            count_max=count_max or count if count is not None else count_max, greedy=greedy
        ).quantifier
        self.raw_pattern = pattern.pattern
        self.pattern = fr"(?P<{name}>{self.raw_pattern}){self.quantifier}"
        self.name = name
        self.alias = pattern.alias
        super(NamedCapturingGroup, self).__init__(self)


class IfEarlierNamedCapturedGroup(PatternType):
    def __init__(self, group_name: NamedCapturingGroup):
        self.pattern = f"(?P={group_name.name})"
        self.alias = group_name.alias
        super(IfEarlierNamedCapturedGroup, self).__init__(self)


class Comment(PatternType):
    def __init__(self, comment_msg: str):
        self.pattern = f"(?#{comment_msg})"
        super(Comment, self).__init__(self)


class StartsWith(PatternType):
    def __init__(self, pattern: PatternType):
        self.pattern = f"^{pattern.pattern}"
        self.alias = pattern.alias
        super(StartsWith, self).__init__(self.pattern)


class EndsWith(PatternType):
    def __init__(self, pattern: PatternType):
        self.pattern = f"{pattern.pattern}$"
        self.alias = pattern.alias
        super(EndsWith, self).__init__(self.pattern)


class StartOfString(PatternType):
    def __init__(self, pattern: PatternType = None):
        self.pattern = fr"\A" if pattern is None else fr"\A{pattern.pattern}"
        self.alias = pattern.alias
        super(StartOfString, self).__init__(self.pattern)


class EndOfString(PatternType):
    def __init__(self, pattern: PatternType = None):
        self.pattern = fr"\Z" if pattern is None else fr"{pattern.pattern}\Z"
        self.alias = pattern.alias
        super(EndOfString, self).__init__(self.pattern)


class WordBoundary(PatternType):
    def __init__(self, pattern: PatternType):
        self.pattern = fr"\b{pattern.pattern}\b"
        self.alias = pattern.alias
        super(WordBoundary, self).__init__(self.pattern)


class IfPrecededBy(PatternType):
    def __init__(self, pattern: PatternType):
        self.pattern = f"(?<={pattern.pattern})"
        self.alias = pattern.alias
        super(IfPrecededBy, self).__init__(self)


class IfNotPrecededBy(PatternType):
    def __init__(self, pattern: PatternType):
        self.pattern = f"(?<!{pattern.pattern})"
        self.alias = pattern.alias
        super(IfNotPrecededBy, self).__init__(self)


class IfFollowedBy(PatternType):
    def __init__(self, pattern: PatternType):
        self.pattern = f"(?={pattern.pattern})"
        self.alias = pattern.alias
        super(IfFollowedBy, self).__init__(self)


class IfNotFollowedBy(PatternType):
    def __init__(self, pattern: PatternType):
        self.pattern = f"(?!{pattern.pattern})"
        self.alias = pattern.alias
        super(IfNotFollowedBy, self).__init__(self)


class IfGroupNameMatched(PatternType):
    def __init__(
            self,
            named_group: NamedCapturingGroup,
            pattern: PatternType,
            fallback_pattern: PatternType = None,
    ):
        self.pattern = (
            f"(?({named_group.name}){pattern.pattern}|{fallback_pattern.pattern})"
            if fallback_pattern is not None
            else f"(?({named_group.name}){pattern.pattern})"
        )
        self.alias = pattern.alias
        super(IfGroupNameMatched, self).__init__(self)


if __name__ == "__main__":
    from pyparsing import Regex

    hyphen = Pattern(r"-", alias="-")
    colon = Pattern(r":", alias=":")
    backslash = Pattern(r"/", alias="/")
    space = Pattern(r"\s", count_min=0, greedy=False, alias=" ")
    four_digits = space & Pattern(r"\d", count=4, alias="dddd") & space
    two_digits = space & Pattern(r"\d", count=2, alias="dd") & space

    # European Date Format
    eu_date_with_hyphen = four_digits & hyphen & two_digits & hyphen & two_digits
    eu_date_with_colon = four_digits & colon & two_digits & colon & two_digits
    eu_date_with_backslash = four_digits & backslash & two_digits & backslash & two_digits
    eu_date = WordBoundary(eu_date_with_colon | eu_date_with_hyphen | eu_date_with_backslash)

    # Indian Date format
    ind_date_with_hyphen = two_digits & hyphen & two_digits & hyphen & four_digits
    ind_date_with_colon = two_digits & colon & two_digits & colon & four_digits
    ind_date_with_backslash = two_digits & backslash & two_digits & backslash & four_digits
    ind_date = WordBoundary(ind_date_with_hyphen | ind_date_with_colon | ind_date_with_backslash)

    start_str = StartOfString(space & Pattern("date") & space & colon & space)
    date_ = (start_str & eu_date) | (start_str & ind_date)

    print(date_.alias)
    print(date_.pattern)
    print(Regex(date_.pattern).re_match("date  : 2021:01:20"))
    print(Regex(date_.pattern).re_match("date: 2021/01/20"))
    print(Regex(date_.pattern).re_match("date : 01-20-2011"))



# Copyright (c) 2022 Valio
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import logging
import os
import pathlib
import sys
import typing
from dataclasses import dataclass

__all__ = ["Logger", "LOGGER", "LOG_LEVEL", "LOG_DIR"]

LOGGER = typing.Union[bool, logging.Logger, None]
LOG_LEVEL = typing.Union[ 
    str, 
    int, 
    typing.List[str], 
    typing.List[int], 
    list[typing.Union[str, int, None]],
    None
    ]
LOG_DIR = typing.Union[str, bytes, pathlib.Path, None]


@dataclass
class Logger(object):
    _stream = False
    _nameToLevel = {
        "DEBUG": logging.DEBUG,
        "INFO": logging.INFO,
        "WARNING": logging.WARNING,
        "ERROR": logging.ERROR,
        "CRITICAL": logging.CRITICAL,
        "FATAL": logging.FATAL,
    }
    _levelToName = {
        logging.DEBUG: "DEBUG",
        logging.INFO: "INFO",
        logging.WARNING: "WARNING",
        logging.ERROR: "ERROR",
        logging.CRITICAL: "CRITICAL",
        logging.FATAL: "FATAL",
    }

    default_format = (
        "%(asctime)-25s:%(module)-15s:%(levelname)-8s:"
        "%(name)-45s:%(funcName)-25s:%(message)s"
    )
    default_log_dir = "logs"
    default_err_format = (
        "%(asctime)-25s:%(module)-15s:%(levelname)-8s:"
        "%(name)-45s:%(funcName)-25s:line-%(lineno)-6s:%(message)s"
    )

    logger: LOGGER = None
    log_levels: LOG_LEVEL = None
    log_dir: LOG_DIR = None

    def __init__(
            self,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
            **kwargs
    ):

        # check logger
        logger_err_msg = TypeError(
            f"expect logger to be a bool or {str(logging.Logger)} type, "
            f"got {type(logger).__name__} type instead"
        )

        if logger is not None and not isinstance(logger, (bool, logging.Logger)):
            raise logger_err_msg

        # check log_levels
        expected_levels = [*self._nameToLevel, *self._levelToName]

        def log_levels_err_msg(unexpected_typ):
            return TypeError(
                f"expect log_levels to be any of {expected_levels} "
                f"or its list, got {unexpected_typ} instead"
            )

        log_levels = (
            [log_levels] if not isinstance(log_levels, (list, tuple)) else log_levels
        )

        unexpected_levels = [
            level
            for level in log_levels
            if not isinstance(level, str) or level not in expected_levels
        ]

        if any(unexpected_levels):
            raise log_levels_err_msg(unexpected_levels)

        # check log_dir
        log_dir_err_msg = TypeError(
            f"expect dir to be a str, bytes or {str(pathlib.Path)} type, "
            f"got {type(log_dir).__name__} type instead"
        )

        if log_dir is not None and not isinstance(log_dir, (str, bytes, pathlib.Path)):
            raise log_dir_err_msg

        # assign
        self.logger = logger
        self.log_levels = log_levels

        curr_dir = os.path.abspath(os.getcwd())
        self.log_dir = (
            os.path.join(
                curr_dir,
                log_dir.decode()
                if isinstance(log_dir, bytes)
                else log_dir or self.default_log_dir,
            )
            if log_dir is not isinstance(log_dir, pathlib.Path)
            else log_dir.joinpath(self.default_log_dir)  # type: ignore
        )
        self._stream = kwargs.pop("stream", self._stream)
        self.kwargs = kwargs

    def get_record_dir(self, folder_name):
        folder_path = (
            os.path.join(self.log_dir, folder_name)
            if not isinstance(self.log_dir, pathlib.Path)
            else self.log_dir.joinpath(folder_name)
        )
        if not isinstance(folder_path, pathlib.Path):
            if not os.path.exists(folder_path):
                os.makedirs(folder_path)
        else:
            folder_path.mkdir(parents=True, exist_ok=True)
        return folder_path

    def get_logger(self, subdir=None, record_prefix=None):
        if self.logger is None or True:
            module_name = os.path.splitext(
                os.path.basename(sys.modules["__main__"].__file__)
            )[0]
            class_name = (
                type(self).__name__
                if subdir is None
                else os.path.splitext(os.path.basename(subdir))[0]
            )
            logger = logging.getLogger(f"{module_name}.{class_name}.{record_prefix}")

            formatter = logging.Formatter(self.default_format)
            err_formatter = logging.Formatter(self.default_err_format)
            logger.setLevel(logging.DEBUG)

            if record_prefix is None:
                record_prefix = class_name

            levels = self.log_levels = (
                self.log_levels if any(self.log_levels) else list(self._nameToLevel)
            )
            levels = [
                level.lower()
                if level not in self._levelToName
                else self._levelToName[level].lower()
                for level in levels
                if level is not None
            ]
            for level in levels:
                lvl_dir = (
                    os.path.join(module_name, level)
                    if subdir is None
                    else os.path.join(os.path.join(module_name, subdir), level)
                )
                level_dir = self.get_record_dir(lvl_dir)
                record_name = f"{record_prefix}.log"
                level_record = (
                    os.path.join(level_dir, record_name)
                    if not isinstance(level_dir, pathlib.Path)
                    else level_dir.join(record_name)
                )
                level_formatter = (
                    formatter if level not in ["critical", "error"] else err_formatter
                )
                level_handler = logging.FileHandler(level_record)
                level_handler.setFormatter(level_formatter)
                level_handler.setLevel(self._nameToLevel[level.upper()])
                logger.addHandler(level_handler)
                if self._stream:
                    stream_handler = logging.StreamHandler()
                    stream_handler.setFormatter(level_formatter)
                    stream_handler.setLevel(self._nameToLevel[level.upper()])
                    logger.addHandler(stream_handler)

            self.logger = logger
        return self.logger



# Copyright (c) 2022 Valio
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT



from typing import Type, Union

from valio import validator as _validator
from valio.descriptor import DEBUG, DEFAULT, DOC, NAME
from valio.error import errors
from valio.logger import LOG_DIR, LOG_LEVEL, LOGGER, loggers
from valio.validator import (BOOL, BYTES, DATE_TIME_DELTA, FLOAT, INT, PATTERN,
                             STR, VALUE, UUID_Type)

# from tortoise import fields


__all__ = [
    "FieldMixin",
    "Field",
    "IntegerField",
    "FloatField",
    "DecimalField",
    "BytesField",
    "BooleanField",
    "StringField",
    "DateField",
    "EmailField",
    "PathField",
    "AadhaarCardField",
    "PaymentCardField",
    "PhoneNumberField",
    "IP4AddressField",
    "IP6AddressField",
    "IPAnyAddressField",
    "UUIDField",
    "EnumField",
    "IntegerEnumField",
    "StringEnumField"
]

_logger = dict()


class FieldBase(loggers.Logger):
    """This class provides base class for all the Field related stuffs"""

    def __set_name__(self, owner, name):
        self.name = name
        _logger[hex(id(self))] = self.get_logger(owner.__name__, name)

    def __class_getitem__(cls, key):
        """getattr"""
        if logger := _logger.get(hex(id(cls)), None):
            logger.info(f"{cls}: getting: {key}")
        try:
            if key not in cls.__dict__:
                raise AttributeError(key)
            return cls.__dict__[key]
        except (errors.GetAttributeError, Exception) as ge:
            if logger:
                logger.error(ge)
            raise ge

    def __setattr__(self, key, value):
        """setattr"""
        if logger := _logger.get(hex(id(self)), None):
            logger.info(f"{self}: setting: {key}")
        try:
            object.__setattr__(self, key, value)
        except (errors.SetAttributeError, Exception) as se:
            if logger:
                logger.error(se)
            raise se

    def __getattr__(self, key):
        """getattr"""
        if logger := _logger.get(hex(id(self)), None):
            logger.info(f"{self}: getting: {key}")
        try:
            if key not in self.__dict__:
                if key not in type(self).__dict__:
                    raise AttributeError(key)
                return type(self).__dict__[key]
            return self.__dict__[key]
        except (errors.GetAttributeError, Exception) as ge:
            if logger:
                logger.error(ge)
            raise ge

    def __delattr__(self, key):
        """delattr"""
        if logger := _logger.get(hex(id(self)), None):
            logger.info(f"{self}: deleting: {key}")
        try:
            del self.__dict__[key]
        except (errors.DeleteAttributeError, Exception) as de:
            if logger:
                logger.error(de)
            raise de

    __getitem__ = __getattr__
    __setitem__ = __setattr__

VALIDATOR = Union[Type[_validator.Validator], _validator.Validator]

class FieldMixin(FieldBase):
    validator: VALIDATOR

    def __init__(
            self,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
    ):
        super(FieldMixin, self).__init__(
            logger=logger,
            log_levels=log_levels,
            log_dir=log_dir,
        )

    def add_validator(self, func):
        self.validator.add_validator(func)
        return func

    def add_pre_validator(self, func):
        self.validator.add_pre_validator(func)
        return func

    def add_post_validator(self, func):
        self.validator.add_post_validator(func)
        return func

    def add_post_set(self, func):
        self.validator.add_post_set(func)
        return func

    def add_pre_get(self, func):
        self.validator.add_pre_get(func)
        return func

    def add_post_get(self, func):
        self.validator.add_post_get(func)
        return func

    def add_pre_delete(self, func):
        self.validator.add_pre_delete(func)
        return func

    def add_post_delete(self, func):
        self.validator.add_post_delete(func)
        return func

    def add_pre_validator_task(self, func):
        self.validator.add_pre_validator_task(func)
        return func

    def add_post_validator_task(self, func):
        self.validator.add_post_validator_task(func)
        return func

    def add_post_set_task(self, func):
        self.validator.add_post_set_task(func)
        return func

    def add_pre_get_task(self, func):
        self.validator.add_pre_get_task(func)
        return func

    def add_post_get_task(self, func):
        self.validator.add_post_get_task(func)
        return func

    def add_pre_delete_task(self, func):
        self.validator.add_pre_delete_task(func)
        return func

    def add_post_delete_task(self, func):
        self.validator.add_post_delete_task(func)
        return func

    def __str__(self):
        return f"{self.validator}"

    def __repr__(self):
        return self.__str__()


class Field(FieldMixin):
    validator: VALIDATOR = _validator.Validator

    def __init__(
            self,
            default: DEFAULT = None,
            name: NAME = None,
            doc: DOC = None,
            required: BOOL = None,
            pattern: PATTERN = None,
            reassign: BOOL = None,
            min_value: VALUE = None,
            value: VALUE = None,
            max_value: VALUE = None,
            min_length: INT = None,
            length: INT = None,
            max_length: INT = None,
            expire_after: DATE_TIME_DELTA = None,
            expire_on: DATE_TIME_DELTA = None,
            expire_before: DATE_TIME_DELTA = None,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
            debug: DEBUG = None,
            enable_async: BOOL = None,
            **kwargs,
    ):
        self.validator = self.validator(
            default=default,
            name=name,
            doc=doc,
            required=required,
            pattern=pattern,
            reassign=reassign,
            min_value=min_value,
            value=value,
            max_value=max_value,
            min_length=min_length,
            length=length,
            max_length=max_length,
            expire_after=expire_after,
            expire_on=expire_on,
            expire_before=expire_before,
            logger=logger,
            log_levels=log_levels,
            log_dir=log_dir,
            debug=debug,
            enable_async=enable_async,
            **kwargs,
        )

        FieldMixin.__init__(self, logger=logger, log_levels=log_levels, log_dir=log_dir)


class IntegerField(FieldMixin):
    validator: VALIDATOR = _validator.IntegerValidator

    def __init__(
            self,
            name: NAME = None,
            default: INT = None,
            required: BOOL = None,
            pattern: PATTERN = None,
            reassign: BOOL = None,
            min_value: INT = None,
            value: INT = None,
            max_value: INT = None,
            min_length: INT = None,
            length: INT = None,
            max_length: INT = None,
            expire_after: DATE_TIME_DELTA = None,
            expire_on: DATE_TIME_DELTA = None,
            expire_before: DATE_TIME_DELTA = None,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
            debug: DEBUG = None,
            enable_async: BOOL = None,
            **kwargs,
    ):
        self.validator = self.validator(
            name=name,
            default=default,
            required=required,
            pattern=pattern,
            reassign=reassign,
            min_value=min_value,
            value=value,
            max_value=max_value,
            min_length=min_length,
            length=length,
            max_length=max_length,
            expire_after=expire_after,
            expire_on=expire_on,
            expire_before=expire_before,
            logger=logger,
            log_levels=log_levels,
            log_dir=log_dir,
            debug=debug,
            enable_async=enable_async,
            **kwargs,
        )

        FieldMixin.__init__(self, logger=logger, log_levels=log_levels, log_dir=log_dir)


class FloatField(FieldMixin):
    validator: VALIDATOR = _validator.FloatValidator

    def __init__(
            self,
            name: NAME = None,
            default: FLOAT = None,
            required: BOOL = None,
            pattern: PATTERN = None,
            reassign: BOOL = None,
            min_value: FLOAT = None,
            value: FLOAT = None,
            max_value: FLOAT = None,
            min_length: INT = None,
            length: INT = None,
            max_length: INT = None,
            expire_after: DATE_TIME_DELTA = None,
            expire_on: DATE_TIME_DELTA = None,
            expire_before: DATE_TIME_DELTA = None,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
            debug: DEBUG = None,
            enable_async: BOOL = None,
            **kwargs,
    ):
        self.validator = self.validator(
            name=name,
            default=default,
            required=required,
            pattern=pattern,
            reassign=reassign,
            min_value=min_value,
            value=value,
            max_value=max_value,
            min_length=min_length,
            length=length,
            max_length=max_length,
            expire_after=expire_after,
            expire_on=expire_on,
            expire_before=expire_before,
            logger=logger,
            log_levels=log_levels,
            log_dir=log_dir,
            debug=debug,
            enable_async=enable_async,
            **kwargs,
        )

        FieldMixin.__init__(self, logger=logger, log_levels=log_levels, log_dir=log_dir)


class DecimalField(FieldMixin):
    validator: VALIDATOR = _validator.DecimalValidator

    def __init__(
            self,
            name: NAME = None,
            default: FLOAT = None,
            required: BOOL = None,
            pattern: PATTERN = None,
            reassign: BOOL = None,
            min_value: FLOAT = None,
            value: FLOAT = None,
            max_value: FLOAT = None,
            min_length: INT = None,
            length: INT = None,
            max_length: INT = None,
            expire_after: DATE_TIME_DELTA = None,
            expire_on: DATE_TIME_DELTA = None,
            expire_before: DATE_TIME_DELTA = None,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
            debug: DEBUG = None,
            enable_async: BOOL = None,
            **kwargs,
    ):
        self.validator = self.validator(
            name=name,
            default=default,
            required=required,
            pattern=pattern,
            reassign=reassign,
            min_value=min_value,
            value=value,
            max_value=max_value,
            min_length=min_length,
            length=length,
            max_length=max_length,
            expire_after=expire_after,
            expire_on=expire_on,
            expire_before=expire_before,
            logger=logger,
            log_levels=log_levels,
            log_dir=log_dir,
            debug=debug,
            enable_async=enable_async,
            **kwargs,
        )

        FieldMixin.__init__(self, logger=logger, log_levels=log_levels, log_dir=log_dir)


class BooleanField(FieldMixin):
    validator: VALIDATOR = _validator.BooleanValidator

    def __init__(
            self,
            name: NAME = None,
            default: BOOL = None,
            required: BOOL = None,
            reassign: BOOL = None,
            expire_after: DATE_TIME_DELTA = None,
            expire_on: DATE_TIME_DELTA = None,
            expire_before: DATE_TIME_DELTA = None,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
            debug: DEBUG = None,
            enable_async: BOOL = None,
            **kwargs,
    ):
        self.validator = self.validator(
            name=name,
            default=default,
            required=required,
            reassign=reassign,
            expire_after=expire_after,
            expire_on=expire_on,
            expire_before=expire_before,
            debug=debug,
            logger=logger,
            log_levels=log_levels,
            log_dir=log_dir,
            enable_async=enable_async,
            **kwargs,
        )

        FieldMixin.__init__(self, logger=logger, log_levels=log_levels, log_dir=log_dir)


class BytesField(FieldMixin):
    validator: VALIDATOR = _validator.BytesValidator

    def __init__(
            self,
            name: NAME = None,
            default: BYTES = None,
            required: BOOL = None,
            pattern: PATTERN = None,
            reassign: BOOL = None,
            min_value: BYTES = None,
            max_value: BYTES = None,
            min_length: INT = None,
            max_length: INT = None,
            expire_after: DATE_TIME_DELTA = None,
            expire_on: DATE_TIME_DELTA = None,
            expire_before: DATE_TIME_DELTA = None,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
            debug: DEBUG = None,
            enable_async: BOOL = None,
            **kwargs,
    ):
        self.validator = self.validator(
            name=name,
            default=default,
            required=required,
            pattern=pattern,
            reassign=reassign,
            min_value=min_value,
            max_value=max_value,
            min_length=min_length,
            max_length=max_length,
            expire_after=expire_after,
            expire_on=expire_on,
            expire_before=expire_before,
            debug=debug,
            logger=logger,
            log_levels=log_levels,
            log_dir=log_dir,
            enable_async=enable_async,
            **kwargs,
        )

        FieldMixin.__init__(self, logger=logger, log_levels=log_levels, log_dir=log_dir)


class StringField(FieldMixin):
    validator: VALIDATOR = _validator.StringValidator

    def __init__(
            self,
            name: NAME = None,
            default: STR = None,
            required: BOOL = None,
            pattern: PATTERN = None,
            reassign: BOOL = None,
            min_value: STR = None,
            value: STR = None,
            max_value: STR = None,
            min_length: INT = None,
            length: INT = None,
            max_length: INT = None,
            expire_after: DATE_TIME_DELTA = None,
            expire_on: DATE_TIME_DELTA = None,
            expire_before: DATE_TIME_DELTA = None,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
            debug: DEBUG = None,
            enable_async: BOOL = None,
            **kwargs,
    ):
        self.validator = self.validator(
            name=name,
            default=default,
            required=required,
            pattern=pattern,
            reassign=reassign,
            min_value=min_value,
            value=value,
            max_value=max_value,
            min_length=min_length,
            length=length,
            max_length=max_length,
            expire_after=expire_after,
            expire_on=expire_on,
            expire_before=expire_before,
            debug=debug,
            logger=logger,
            log_levels=log_levels,
            log_dir=log_dir,
            enable_async=enable_async,
            **kwargs,
        )

        FieldMixin.__init__(self, logger=logger, log_levels=log_levels, log_dir=log_dir)


class DateField(FieldMixin):
    validator: VALIDATOR = _validator.DateValidator

    def __init__(
            self,
            name: NAME = None,
            default: DATE_TIME_DELTA = None,
            required: BOOL = None,
            reassign: BOOL = None,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
            debug: DEBUG = None,
            enable_async: BOOL = None,
            **kwargs,
    ):
        self.validator = self.validator(
            name=name,
            default=default,
            required=required,
            reassign=reassign,
            debug=debug,
            logger=logger,
            log_levels=log_levels,
            log_dir=log_dir,
            enable_async=enable_async,
            **kwargs
        )

        FieldMixin.__init__(self, logger=logger, log_levels=log_levels, log_dir=log_dir)


class PathField(FieldMixin):
    validator: VALIDATOR = _validator.PathValidator

    def __init__(
            self,
            default=None,
            required: BOOL = None,
            reassign: BOOL = None,
            pattern: PATTERN = None,
            min_length: INT = None,
            length: INT = None,
            max_length: INT = None,
            expire_after: DATE_TIME_DELTA = None,
            expire_on: DATE_TIME_DELTA = None,
            expire_before: DATE_TIME_DELTA = None,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
            name: NAME = None,
            doc: DOC = None,
            debug: DEBUG = None,
            **kwargs,
    ):
        self.validator = self.validator(
            default=default,
            required=required,
            reassign=reassign,
            pattern=pattern,
            min_length=min_length,
            length=length,
            max_length=max_length,
            expire_after=expire_after,
            expire_on=expire_on,
            expire_before=expire_before,
            name=name,
            doc=doc,
            debug=debug,
            **kwargs,
        )

        FieldMixin.__init__(self, logger=logger, log_levels=log_levels, log_dir=log_dir)


class IP4AddressField(FieldMixin):
    validator: VALIDATOR = _validator.IP4AddressValidator

    def __init__(
            self,
            default=None,
            required: BOOL = None,
            reassign: BOOL = None,
            pattern: PATTERN = None,
            min_value: STR = None,
            value: STR = None,
            max_value: STR = None,
            min_length: INT = None,
            length: INT = None,
            max_length: INT = None,
            expire_after: DATE_TIME_DELTA = None,
            expire_on: DATE_TIME_DELTA = None,
            expire_before: DATE_TIME_DELTA = None,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
            name: NAME = None,
            doc: DOC = None,
            debug: DEBUG = None,
            **kwargs,
    ):
        self.validator = self.validator(
            default=default,
            required=required,
            reassign=reassign,
            pattern=pattern,
            min_value=min_value,
            value=value,
            max_value=max_value,
            min_length=min_length,
            length=length,
            max_length=max_length,
            expire_after=expire_after,
            expire_on=expire_on,
            expire_before=expire_before,
            name=name,
            doc=doc,
            debug=debug,
            **kwargs,
        )

        FieldMixin.__init__(self, logger=logger, log_levels=log_levels, log_dir=log_dir)


class IP6AddressField(FieldMixin):
    validator: VALIDATOR = _validator.IP6AddressValidator

    def __init__(
            self,
            default=None,
            required: BOOL = None,
            reassign: BOOL = None,
            pattern: PATTERN = None,
            min_value: STR = None,
            value: STR = None,
            max_value: STR = None,
            min_length: INT = None,
            length: INT = None,
            max_length: INT = None,
            expire_after: DATE_TIME_DELTA = None,
            expire_on: DATE_TIME_DELTA = None,
            expire_before: DATE_TIME_DELTA = None,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
            name: NAME = None,
            doc: DOC = None,
            debug: DEBUG = None,
            **kwargs,
    ):
        self.validator = self.validator(
            default=default,
            required=required,
            reassign=reassign,
            pattern=pattern,
            min_value=min_value,
            value=value,
            max_value=max_value,
            min_length=min_length,
            length=length,
            max_length=max_length,
            expire_after=expire_after,
            expire_on=expire_on,
            expire_before=expire_before,
            name=name,
            doc=doc,
            debug=debug,
            **kwargs,
        )

        FieldMixin.__init__(self, logger=logger, log_levels=log_levels, log_dir=log_dir)


class IPAnyAddressField(FieldMixin):
    validator: VALIDATOR = _validator.IPAnyAddressValidator

    def __init__(
            self,
            default=None,
            required: BOOL = None,
            reassign: BOOL = None,
            pattern: PATTERN = None,
            min_value: STR = None,
            value: STR = None,
            max_value: STR = None,
            min_length: INT = None,
            length: INT = None,
            max_length: INT = None,
            expire_after: DATE_TIME_DELTA = None,
            expire_on: DATE_TIME_DELTA = None,
            expire_before: DATE_TIME_DELTA = None,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
            name: NAME = None,
            doc: DOC = None,
            debug: DEBUG = None,
            **kwargs,
    ):
        self.validator = self.validator(
            default=default,
            required=required,
            reassign=reassign,
            pattern=pattern,
            min_value=min_value,
            value=value,
            max_value=max_value,
            min_length=min_length,
            length=length,
            max_length=max_length,
            expire_after=expire_after,
            expire_on=expire_on,
            expire_before=expire_before,
            name=name,
            doc=doc,
            debug=debug,
            **kwargs,
        )

        FieldMixin.__init__(self, logger=logger, log_levels=log_levels, log_dir=log_dir)


class EmailField(FieldMixin):
    validator: VALIDATOR = _validator.EmailIDValidator

    def __init__(
            self,
            default=None,
            required: BOOL = None,
            reassign: BOOL = None,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
            name: NAME = None,
            doc: DOC = None,
            debug: DEBUG = None,
            **kwargs
    ):
        self.validator = self.validator(
            default=default,
            required=required,
            reassign=reassign,
            name=name,
            doc=doc,
            debug=debug,
            logger=logger,
            log_levels=log_levels,
            log_dir=log_dir,
            **kwargs
        )

        FieldMixin.__init__(self, logger=logger, log_levels=log_levels, log_dir=log_dir)


class AadhaarCardField(FieldMixin):
    validator: VALIDATOR = _validator.AadhaarCardValidator

    def __init__(
            self,
            default=None,
            required: BOOL = None,
            reassign: BOOL = None,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
            name: NAME = None,
            doc: DOC = None,
            debug: DEBUG = None,
            **kwargs
    ):
        self.validator = self.validator(
            default=default,
            required=required,
            reassign=reassign,
            name=name,
            doc=doc,
            debug=debug,
            logger=logger,
            log_levels=log_levels,
            log_dir=log_dir,
            **kwargs
        )

        FieldMixin.__init__(self, logger=logger, log_levels=log_levels, log_dir=log_dir)


class PaymentCardField(FieldMixin):
    validator: VALIDATOR = _validator.PaymentCardValidator

    def __init__(
            self,
            default=None,
            required: BOOL = None,
            reassign: BOOL = None,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
            name: NAME = None,
            doc: DOC = None,
            debug: DEBUG = None,
            **kwargs
    ):
        self.validator = self.validator(
            default=default,
            required=required,
            reassign=reassign,
            name=name,
            doc=doc,
            debug=debug,
            logger=logger,
            log_levels=log_levels,
            log_dir=log_dir,
            **kwargs
        )

        FieldMixin.__init__(self, logger=logger, log_levels=log_levels, log_dir=log_dir)


class PhoneNumberField(FieldMixin):
    validator: VALIDATOR = _validator.PhoneNumberValidator

    def __init__(
            self,
            default=None,
            required: BOOL = None,
            reassign: BOOL = None,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
            name: NAME = None,
            doc: DOC = None,
            debug: DEBUG = None,
            **kwargs
    ):
        self.validator = self.validator(
            default=default,
            required=required,
            reassign=reassign,
            name=name,
            doc=doc,
            debug=debug,
            logger=logger,
            log_levels=log_levels,
            log_dir=log_dir,
            **kwargs
        )

        FieldMixin.__init__(self, logger=logger, log_levels=log_levels, log_dir=log_dir)


class UUIDField(FieldMixin):
    validator: VALIDATOR = _validator.UUIDValidator

    def __init__(
            self,
            name: NAME = None,
            default: UUID_Type = None,
            required: BOOL = None,
            reassign: BOOL = None,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
            debug: DEBUG = None,
            enable_async: BOOL = None,
            **kwargs,
    ):
        self.validator = self.validator(
            name=name,
            default=default,
            required=required,
            reassign=reassign,
            debug=debug,
            logger=logger,
            log_levels=log_levels,
            log_dir=log_dir,
            enable_async=enable_async,
            **kwargs
        )

        FieldMixin.__init__(self, logger=logger, log_levels=log_levels, log_dir=log_dir)


class EnumField(FieldMixin):
    validator: VALIDATOR = _validator.EnumValidator

    def __init__(
            self,
            default=None,
            required: BOOL = None,
            reassign: BOOL = None,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
            name: NAME = None,
            doc: DOC = None,
            debug: DEBUG = None,
            **kwargs
    ):
        self.validator = self.validator(
            default=default,
            required=required,
            reassign=reassign,
            name=name,
            doc=doc,
            debug=debug,
            logger=logger,
            log_levels=log_levels,
            log_dir=log_dir,
            **kwargs
        )

        FieldMixin.__init__(self, logger=logger, log_levels=log_levels, log_dir=log_dir)


class IntegerEnumField(FieldMixin):
    validator: VALIDATOR = _validator.IntegerEnumValidator

    def __init__(
            self,
            default=None,
            required: BOOL = None,
            reassign: BOOL = None,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
            name: NAME = None,
            doc: DOC = None,
            debug: DEBUG = None,
            **kwargs
    ):
        self.validator = self.validator(
            default=default,
            required=required,
            reassign=reassign,
            name=name,
            doc=doc,
            debug=debug,
            logger=logger,
            log_levels=log_levels,
            log_dir=log_dir,
            **kwargs
        )

        FieldMixin.__init__(self, logger=logger, log_levels=log_levels, log_dir=log_dir)


class StringEnumField(FieldMixin):
    validator: VALIDATOR = _validator.StringEnumValidator

    def __init__(
            self,
            default=None,
            required: BOOL = None,
            reassign: BOOL = None,
            logger: LOGGER = None,
            log_levels: LOG_LEVEL = None,
            log_dir: LOG_DIR = None,
            name: NAME = None,
            doc: DOC = None,
            debug: DEBUG = None,
            **kwargs
    ):
        self.validator = self.validator(
            default=default,
            required=required,
            reassign=reassign,
            name=name,
            doc=doc,
            debug=debug,
            logger=logger,
            log_levels=log_levels,
            log_dir=log_dir,
            **kwargs
        )

        FieldMixin.__init__(self, logger=logger, log_levels=log_levels, log_dir=log_dir)
